# Copyright 2021 MosaicML. All Rights Reserved.

from composer.algorithms.colout.colout import ColOut as ColOut
from composer.algorithms.colout.colout import ColOutHparams as ColOutHparams
from composer.algorithms.colout.colout import colout as colout

_name = 'ColumnOut'
_class_name = 'ColOut'
_functional = 'colout'
_tldr = 'Removes columns and rows from the image for augmentation and efficiency.'
_attribution = 'MosaicML'
_link = ''
_method_card = ''



# Copyright 2021 MosaicML. All Rights Reserved.

from composer.algorithms.randaugment.randaugment import RandAugment as RandAugment
from composer.algorithms.randaugment.randaugment import RandAugmentHparams as RandAugmentHparams
from composer.algorithms.randaugment.randaugment import RandAugmentTransform as RandAugmentTransform
from composer.algorithms.randaugment.randaugment import randaugment as randaugment

_name = 'RandAugment'
_class_name = 'RandAugment'
_functional = 'randaugment'
_tldr = 'Applies a series of random augmentations'
_attribution = '(Cubuk et al, 2020)'
_link = 'https://openaccess.thecvf.com/content_CVPRW_2020/html/w40/Cubuk_Randaugment_Practical_Automated_Data_Augmentation_With_a_Reduced_Search_Space_CVPRW_2020_paper.html'
_method_card = ''



# Copyright 2021 MosaicML. All Rights Reserved.

from __future__ import annotations

import warnings
from enum import Enum


class StringEnum(Enum):
    __hash__ = Enum.__hash__

    def __eq__(self, o: object) -> bool:
        if isinstance(o, str):
            cls_name = self.__class__.__name__
            warnings.warn(
                f"Detected comparision between a string and {cls_name}. Please use {cls_name}({o}) "
                f"to convert both types to {cls_name} before comparing",
                category=UserWarning)
            o_enum = type(self)(o)
            return super().__eq__(o_enum)
        return super().__eq__(o)

    def __init__(self, *args: object) -> None:
        if self.name.upper() != self.name:
            raise ValueError(
                f"{self.__class__.__name__}.{self.name} is invalid. All keys in {self.__class__.__name__} must be uppercase. "
                f"To fix, rename to '{self.name.upper()}'.")
        if self.name.lower() != self.value:
            raise ValueError(f"{self.__class__.__name__}.{self.name} has an invalid value {self.value}. "
                             f"The value must be the lowercase value as its key, {self.name}. "
                             f"To fix, rename the value to '{self.name.lower()}''.")

    @classmethod
    def _missing_(cls, value: object) -> StringEnum:
        # Override _missing_ so both lowercase and uppercase names are supported,
        # as well as passing an instance through
        if isinstance(value, cls):
            return value
        if isinstance(value, str):
            try:
                return cls[value.upper()]
            except KeyError as e:
                raise ValueError(f"{cls.__name__} has no key for {value}") from e
        raise TypeError(f"Unable to convert value({value}) of type {type(value)} into {cls.__name__}")



# Copyright 2021 MosaicML. All Rights Reserved.

from __future__ import annotations

import collections.abc
import operator
from copy import deepcopy
from enum import IntEnum
from functools import reduce
from typing import TYPE_CHECKING, Callable, Generator, Mapping, Sequence, Union

import torch

if TYPE_CHECKING:
    from composer.core.logging.base_backend import BaseLoggerBackend
    from composer.core.state import State

TLogDataValue = Union[str, float, int, torch.Tensor, Sequence["TLogDataValue"], Mapping[str, "TLogDataValue"]]
TLogData = Mapping[str, TLogDataValue]


class LogLevel(IntEnum):
    """LogLevel denotes where in the training loop log messages are generated.

    Logging destinations use the LogLevel to determine whether to record a given
    metric or state change
    """
    FIT = 1
    EPOCH = 2
    BATCH = 3
    MICROBATCH = 4
    VERBOSE = 5


class Logger:
    """
    Logger records metrics and state changes to logging destinations.

    It routes logging calls to the
    :class:`~compose.core.logging.base_backend.BaseLoggerBackend`s
    (specified in :param log_destinations:)

    Args:
        state (State): The global :class:`~composer.core.State` object.
        log_destinations (Sequence[BaseLoggerBackend]):
            A sequence of
            :class:`~compose.core.logging.base_backend.BaseLoggerBackend`s
            to which logging calls will be sent.
    """

    def __init__(
            self,
            state: State,
            log_destinations: Sequence[BaseLoggerBackend] = tuple(),  # a sequence of logging destinations
    ):
        # destinations are constructed on first use once we have the state
        self._log_destinations = log_destinations

        self._state = state

    def _get_destinations_for_log_level(self, log_level: LogLevel) -> Generator[BaseLoggerBackend, None, None]:
        for destination in self._log_destinations:
            if destination.will_log(self._state, log_level):
                yield destination

    def metric(self, log_level: Union[str, LogLevel], data: Union[TLogData, Callable[[], TLogData]]) -> None:
        """
        Send :param data: with :param log_level: to the logging backends.

        Args:
            log_level (Union[str, LogLevel]): A :class:`LogLevel`.
            data (Union[TLogData, Callable[[], TLogData]]):
                Can be either logging data or a callable that returns
                data to be logged. Callables will be invoked
                only when :meth:`will_log` returns True for at least one 
                :class:`~compose.core.logging.base_backend.BaseLoggerBackend`.
                Useful when it is expensive to generate the data to be logged.
        """
        if isinstance(log_level, str):
            log_level = LogLevel[log_level.upper()]

        for destination in self._get_destinations_for_log_level(log_level):
            if callable(data):
                data = data()
            # copying the data in case if a backend queues the logged data and flushes later
            # this way, the flushed data will be the same as at the time of the logger call
            copied_data = deepcopy(data)
            assert isinstance(copied_data, collections.abc.Mapping)
            destination.log_metric(self._state.epoch, self._state.step, log_level, copied_data)

    def metric_fit(self, data: Union[TLogData, Callable[[], TLogData]]) -> None:
        """Helper function for `self.metric(LogLevel.FIT, data)`"""
        self.metric(LogLevel.FIT, data)

    def metric_epoch(self, data: Union[TLogData, Callable[[], TLogData]]) -> None:
        """Helper function for `self.metric(LogLevel.EPOCH, data)`"""
        self.metric(LogLevel.EPOCH, data)

    def metric_batch(self, data: Union[TLogData, Callable[[], TLogData]]) -> None:
        """Helper function for `self.metric(LogLevel.BATCH, data)`"""
        self.metric(LogLevel.BATCH, data)

    def metric_microbatch(self, data: Union[TLogData, Callable[[], TLogData]]) -> None:
        """Helper function for `self.metric(LogLevel.MICROBATCH, data)`"""
        self.metric(LogLevel.MICROBATCH, data)

    def metric_verbose(self, data: Union[TLogData, Callable[[], TLogData]]) -> None:
        """Helper function for `self.metric(LogLevel.VERBOSE, data)`"""
        self.metric(LogLevel.VERBOSE, data)


def format_log_data_value(data: TLogDataValue) -> str:
    if data is None:
        return "None"
    if isinstance(data, str):
        return f"\"{data}\""
    if isinstance(data, int):
        return str(data)
    if isinstance(data, float):
        return f"{data:.4f}"
    if isinstance(data, torch.Tensor):
        if data.shape == tuple() or reduce(operator.mul, data.shape, 1) == 1:
            return format_log_data_value(data.cpu().item())
        return "Tensor of shape " + str(data.shape)
    if isinstance(data, collections.abc.Mapping):
        output = ['{ ']
        for k, v in data.items():
            assert isinstance(k, str)
            v = format_log_data_value(v)
            output.append(f"\"{k}\": {v}, ")
        output.append('}')
        return "".join(output)
    if isinstance(data, collections.abc.Iterable):
        return "[" + ", ".join(format_log_data_value(v) for v in data) + "]"
    raise NotImplementedError(f"Unable to format variable of type: {type(data)} with value {data}")



# Copyright 2021 MosaicML. All Rights Reserved.

from composer.algorithms.scale_schedule.scale_schedule import ScaleSchedule as ScaleSchedule
from composer.algorithms.scale_schedule.scale_schedule import ScaleScheduleHparams as ScaleScheduleHparams
from composer.algorithms.scale_schedule.scale_schedule import scale_scheduler as scale_scheduler

_name = 'Scale Schedule'
_class_name = 'ScaleScheduler'
_functional = 'scale_scheduler'
_tldr = 'Scales the learning schedule by a factor'
_attribution = ''
_link = ''
_method_card = ''



# Copyright 2021 MosaicML. All Rights Reserved.

from composer.algorithms.mixup.mixup import MixUp as MixUp
from composer.algorithms.mixup.mixup import MixUpHparams as MixUpHparams
from composer.algorithms.mixup.mixup import mixup_batch as mixup_batch

_name = 'MixUp'
_class_name = 'MixUp'
_functional = 'mixup_batch'
_tldr = 'Blends pairs of examples and labels'
_attribution = '(Zhang et al, 2017)'
_link = 'https://arxiv.org/abs/1710.09412'
_method_card = ''



# Copyright 2021 MosaicML. All Rights Reserved.

import math
from dataclasses import asdict, dataclass
from typing import Dict, Mapping, Optional

import torch
import yahp as hp

from composer.algorithms import AlgorithmHparams
from composer.core.types import Algorithm, Batch, Event, Logger, State, Tensor
from composer.models.transformer_shared import MosaicTransformer
from composer.utils import ensure_tuple


def apply_curriculum(batch: Dict[str, Tensor], curr_seq_len: int, truncate: bool) -> Batch:
    """
    Progressively increases the sequence length during training.

    Changes the sequence length of all tensors in the provided dictionary
    to ``curr_seq_len``, by either truncating the tensors (``truncate=True``)
    or reshaping the tensors to create new examples from the extra tokens
    (``truncate=False``).

    The schedule for ``curr_seq_len`` over training time should be managed
    out of this function.

    Args:
        batch: The input batch to the model, must be a dictionary.
        curr_seq_length (int): The desired sequence length to apply.
        truncate (bool): Truncate sequences early, or reshape tensors
                         to create new examples out of the extra tokens.

    Returns:
        batch: a Mapping of input tensors to the model,
               where all tensors have curr_seq_len in the second dimension.
    """

    assert isinstance(batch, Mapping)

    if truncate:
        for k in batch.keys():
            batch[k] = batch[k][:, :curr_seq_len]
    else:
        # ensure new tensor shape is divisible by curr_seq_len
        input_ids = batch['input_ids'].view(-1)
        tensor_len = (input_ids.shape[0] // curr_seq_len) * curr_seq_len

        input_ids = input_ids[:tensor_len]
        input_ids = input_ids.view(-1, curr_seq_len)
        batch['input_ids'] = input_ids

        for k, v in batch.items():
            if k == "input_ids":
                continue
            v = v.view(-1)
            v = v[:tensor_len]
            batch[k] = v.view(-1, curr_seq_len)

    return batch


@dataclass
class CurriculumLearningHparams(AlgorithmHparams):

    duration: float = hp.optional("Fraction of total training time to apply curriculum learning.", default=0.3)
    min_seq_length: int = hp.optional("Starting sequence length.", default=8)
    max_seq_length: int = hp.optional("End sequence length", default=1024)
    step_size: int = hp.optional("Sequence length step size", default=8)
    truncate: bool = hp.optional("Truncate tensors or reshape extra tokens to new examples.", default=True)

    def validate(self):
        if self.duration < 0 or self.duration > 1:
            raise ValueError(f'Duration must be getween 0 and 1, got: {self.duration}')

        if self.max_seq_length < self.min_seq_length:
            raise ValueError(f'max_seq_length={self.max_seq_length} must be '
                             f'greater than min_seq_length={self.min_seq_length}')

    def initialize_object(self) -> "CurriculumLearning":
        return CurriculumLearning(**asdict(self))


class CurriculumLearning(Algorithm):
    """Progressively increases the sequence length during training.

    Changes the sequence length of all tensors in the input batch. The
    sequence length increases from ``min_seq_length`` to ``max_seq_length``
    in steps of ``step_size`` during the first ``duration`` fraction of
    training.

    The sequence length is then kept at ``max_seq_length``
    for the rest of training.

    Tensors are either truncated (``truncate=True``) or reshaped to
    create new examples from the extra tokens (``truncate=False``).

    .. note::

        ``step_size`` should be a multiple of eight for GPUs

    .. note::

        Variable input lengths can create CUDA OOM errors. To avoid this,
        we follow pytorch notes and pre-allocate the memory with a blank
        forward and backward pass.

    Args:
        duration (float): fraction of total training for curriculum learning.
        min_seq_length (int): Minimum sequence length to start the curriculum.
        max_seq_length (int): Maximum sequence length to stop the curriculum.
        step_size (int): Step size of sequence length.

        truncate (bool): Truncate tensors or reshape extra tokens to new examples
    """

    def __init__(
        self,
        duration: float = 0.3,
        min_seq_length: int = 8,
        max_seq_length: int = 1024,
        step_size: int = 8,
        truncate: bool = True,
    ):
        self.hparams = CurriculumLearningHparams(duration=duration,
                                                 min_seq_length=min_seq_length,
                                                 max_seq_length=max_seq_length,
                                                 step_size=step_size,
                                                 truncate=truncate)
        self.hparams.validate()

    def match(self, event: Event, state: State) -> bool:
        return event in (Event.TRAINING_START, Event.AFTER_DATALOADER)

    def apply(self, event: Event, state: State, logger: Logger) -> Optional[int]:

        # in order to avoid OOMs, we do a forward and a backward pass on a dummy input.
        if event == Event.TRAINING_START:
            # ensure that input_ids is a valid model input. since we don't need the
            # results, we don't use all inputs.

            original_model = state.model.module
            assert isinstance(original_model, MosaicTransformer)
            model_inputs = original_model.get_model_inputs()  # type: ignore
            assert 'input_ids' in model_inputs
            assert 'labels' in model_inputs

            # create fake inputs
            vocab_size = len(original_model.tokenizer)  # type: ignore

            # simplifying assumption: Composer doesn't support model-parallelism,
            # so the first parameter's device is likely the same device for
            # all of the parameters
            device = next(state.model.parameters()).device

            assert (state.train_batch_size % state.world_size) == 0
            per_gpu_batch = math.ceil(state.train_batch_size / (state.world_size * state.grad_accum))
            input_ids = torch.randint(low=0,
                                      high=vocab_size - 1,
                                      size=(per_gpu_batch, self.hparams.max_seq_length),
                                      device=device).long()
            labels = input_ids.clone()
            attn_mask = torch.ones_like(labels)
            model_inputs = {
                "input_ids": input_ids,
                "labels": labels,
                "attention_mask": attn_mask,
            }

            # start by running a forward and backward pass
            # of the maximum sequence length to allocate cache.
            with state.precision_context(state.precision):
                outputs = state.model.forward(model_inputs)
                loss = original_model.loss(outputs, model_inputs)

            # since use_grad_scaling is in the Trainer, and we
            # don't care about the loss values, skip scaling
            for loss_item in ensure_tuple(loss):
                loss_item.backward()

            # zero out gradients and proceed to normal training
            assert state.optimizers is not None, \
                "optimizers are set before TRAINING_START"

            for optimizer in ensure_tuple(state.optimizers):
                optimizer.zero_grad()
        else:
            num_optimization_steps = state.steps_per_epoch * state.max_epochs
            num_curriculum_steps = int(num_optimization_steps * self.hparams.duration)

            # assume the full sequence length is the unaltered sequence length
            num_update_steps = (self.hparams.max_seq_length - self.hparams.min_seq_length) // self.hparams.step_size
            update_every_n_steps = num_curriculum_steps // num_update_steps

            curr_seq_len = self.hparams.step_size * (state.step // update_every_n_steps)
            curr_seq_len = max(curr_seq_len, self.hparams.min_seq_length)
            curr_seq_len = min(curr_seq_len, self.hparams.max_seq_length)

            state.batch = apply_curriculum(state.batch_dict, curr_seq_len, self.hparams.truncate)

            batch_size = state.batch_dict['input_ids'].shape[0]
            logger.metric_batch({
                'curriculum_learning/curr_seq_len': curr_seq_len,
                'curriculum_learning/curr_bs': batch_size,
            })



# Copyright 2021 MosaicML. All Rights Reserved.

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, Callable, List, NamedTuple, Optional, Sequence

import torch
import yahp as hp

from composer.core.types import Batch, Dataset, Tensor, TPrefetchFn


def _split_fn(batch: Batch, n_microbatches: int) -> List[Batch]:
    if not isinstance(batch, Sequence):
        raise ValueError(f'split_fn requires batch be a tuple pair of tensors, got {type(batch)}')
    x, y = batch
    if isinstance(x, Tensor) and isinstance(y, Tensor):
        return list(zip(x.chunk(n_microbatches), y.chunk(n_microbatches)))
    if isinstance(x, List) and isinstance(y, List):
        return list(
            zip(
                [x[i::n_microbatches] for i in range(n_microbatches)],
                [y[i::n_microbatches] for i in range(n_microbatches)],
            ))
    raise NotImplementedError('The default split_fn is unable to split the output of this'
                              'dataloader. Please define a split_fn in your dataloader spec.')


class DataloaderSpec(NamedTuple):
    dataset: Dataset
    drop_last: bool
    shuffle: bool
    collate_fn: Optional[Callable[[List[Any]], Batch]] = None
    worker_init_fn: Optional[Callable[[int], None]] = None
    multiprocessing_context: Any = None
    generator: Optional[torch.Generator] = None
    prefetch_fn: Optional[TPrefetchFn] = None
    split_fn: Callable[[Batch, int], List[Batch]] = _split_fn


@dataclass
class DatasetHparams(hp.Hparams, ABC):
    pass

    @abstractmethod
    def initialize_object(self) -> DataloaderSpec:
        pass



# Copyright 2021 MosaicML. All Rights Reserved.

# type: ignore
from __future__ import annotations

import logging
from typing import Optional

import torch
from torch.optim.lr_scheduler import CosineAnnealingLR
from torch.optim.swa_utils import SWALR, AveragedModel

from composer.algorithms.swa.hparams import SWAHparams
from composer.core.types import Algorithm, Event, Logger, State

log = logging.getLogger(__name__)

import math

import torch
from torch.nn import Module


@torch.no_grad()
def update_bn(loader, model, device=None):
    """
    Updates BatchNorm running_mean, running_var buffers in the model.
    It performs one pass over data in `loader` to estimate the activation
    statistics for BatchNorm layers in the model.
    Adapted from https://github.com/pytorch/pytorch/blob/master/torch/optim/swa_utils.py
    in order to work with our internal trainer.
    """
    momenta = {}
    for module in model.modules():
        if isinstance(module, torch.nn.modules.batchnorm._BatchNorm):
            module.running_mean = torch.zeros_like(module.running_mean)
            module.running_var = torch.ones_like(module.running_var)
            momenta[module] = module.momentum

    if not momenta:
        return

    was_training = model.training
    model.train()
    for module in momenta.keys():
        module.momentum = None
        module.num_batches_tracked *= 0

    for i, data in enumerate(loader):
        model(data)

    for bn_module in momenta.keys():
        bn_module.momentum = momenta[bn_module]
    model.train(was_training)


class SWA(Algorithm):
    """Apply Stochastic Weight Averaging

    Stochastic Weight Averaging (SWA) averages model weights sampled towards the end of training.
    This leads to better generalization than conventional training.

    See `Averaging Weights Leads to Wider Optima and Better Generalization` <https://arxiv.org/abs/1803.05407>.

    
    Args:
        swa_start (float): fraction of training completed before stochastic weight averaging is applied
        anneal_epochs (int): The final learning rate to anneal towards
        swa_lr (float): fraction of minibatch to select and keep for gradient computation

    """

    def __init__(self, swa_start: float = 0.8, anneal_epochs: int = 10, swa_lr: Optional[float] = None):
        self.hparams = SWAHparams(
            swa_start=swa_start,
            anneal_epochs=anneal_epochs,
            swa_lr=swa_lr,
        )
        assert 0 < swa_start < 1, "swa_start must be between 0 and 1."
        assert anneal_epochs > 0, "anneal_epochs must be great than 0."

        self.swa_scheduler = None

    def match(self, event: Event, state: State) -> bool:
        should_start_swa = state.epoch >= int(self.hparams.swa_start * state.max_epochs)
        return event in (Event.TRAINING_START, Event.TRAINING_END) or \
             (event == Event.EPOCH_END and should_start_swa)

    def apply(self, event: Event, state: State, logger: Logger) -> None:

        assert state.model is not None, 'We cannot apply SWA to None'

        swa_start_epochs = int(self.hparams.swa_start * state.max_epochs)

        if event == Event.TRAINING_START:
            self.swa_model = AveragedModel(state.model)

        if event == Event.EPOCH_END and state.epoch == swa_start_epochs:
            assert self.swa_scheduler is None, "SWA Scheduler should only be set once. Another algorithm "
            "may have adjusted the max_epochs."

            if self.hparams.swa_lr is None:
                last_lr = state.schedulers.schedulers[0].get_last_lr()  # assumes ComposedScheduler
                log.info(f'Setting SWA LR to {last_lr}')
                self.hparams.swa_lr = last_lr

            self.swa_scheduler = SWALR(
                state.optimizers[0] if isinstance(state.optimizers, tuple) else state.optimizers,
                swa_lr=self.hparams.swa_lr,
                anneal_epochs=self.hparams.anneal_epochs,
                anneal_strategy='cos',
            )

        if event == Event.EPOCH_END and state.epoch >= swa_start_epochs:
            self.swa_model.update_parameters(state.model)

            if self.swa_scheduler is None:
                raise ValueError('SWA LR scheduler was not set.')
            self.swa_scheduler.step()

        ## end of training
        if event == Event.TRAINING_END:
            update_bn(state.train_dataloader, self.swa_model)
            state.model = self.swa_model
            log.info('Updated BN and set model to the averaged model')



# Copyright 2021 MosaicML. All Rights Reserved.

"""
Helper function to generate the README table

"""
import os
from pathlib import Path

from composer import utils

HEADER = ['Task', 'Dataset', 'Name', 'Quality', 'Metric', 'TTT', 'Hparams']
ATTRIBUTES = ['_task', '_dataset', '_name', '_quality', '_metric', '_ttt', '_hparams']
GITHUB_BASE = 'https://github.com/mosaicml/mosaicml/tree/dev/composer/models/'  # TODO: update for launch

folder_path = Path(__file__).parent
models = utils.list_dirs(folder_path)

if not len(models):
    raise ValueError(f'Found 0 models in {folder_path}')

print(f'Found {len(models)} models')

metadata = utils.get_metadata(
    names=models,
    attributes=ATTRIBUTES,
    module_basepath='composer.models',
)

# add extra keys
for name, md in metadata.items():
    md['_github_link'] = GITHUB_BASE + name
    md['_hparams_path'] = os.path.join('composer', 'models', name, md['_hparams'])
    md['_hparams_link'] = f"{GITHUB_BASE}/{name}/{md['_hparams']}"

# define row format
row = [
    '{_task}',
    '{_dataset}',
    '[{_name}]({_github_link})',
    '{_quality}',
    '{_metric}',
    '{_ttt}',
    '[{_hparams_path}]({_hparams})',
]

table_md = utils.build_markdown_table(
    header=HEADER,
    metadata=metadata,
    sorted_keys=sorted(metadata.keys()),
    row_format=row,
)

# update table in README.md
source_file = Path(__file__).parent.joinpath('README.md')
utils.update_table_in_file(table_md, source_file)



# Copyright 2021 MosaicML. All Rights Reserved.

from __future__ import annotations

import logging
import warnings
from dataclasses import dataclass, field, fields
from typing import TYPE_CHECKING, Callable, ContextManager, Optional, Sequence, Union

import torch.nn.modules.utils
from torch.nn.parallel import DistributedDataParallel

import composer.core.types as types
from composer.core.callback import Callback
from composer.core.precision import Precision
from composer.core.serializable import Serializable
from composer.utils import ensure_tuple, make_empty_tensor
from composer.utils.ddp import get_global_rank, is_rank_set
from composer.utils.precision import default_precision_factory

if TYPE_CHECKING:
    from composer.core.algorithm import Algorithm

logger = logging.getLogger(__name__)

# These fields will be serialized directly using torch.save / torch.load
DIRECT_SERIALIZATION_FIELDS = [
    "train_batch_size",
    "eval_batch_size",
    "last_batch_size",
    "grad_accum",
    "_precision",
    "max_epochs",
    "epoch",
    "step",
    "seed",
]

# These fields will be serialized using .state_dict(), and loaded with .load_state_dict()
STATE_DICT_SERIALIZATION_FIELDS = [
    "model",
    "optimizers",
    "schedulers",
    "algorithms",
    "callbacks",
    "scaler",
]

# These fields will not be serialized
SKIP_SERIALIZATION_FIELDS = [
    "loss",
    "batch",
    "outputs",
    "precision",
    "train_dataloader",
    "eval_dataloader",
    "world_size",
    "nproc_per_node",
    "precision",
    "precision_context",
]


@dataclass
class State(Serializable):
    """
    The current state of the trainer.

    Algorithms are able to modify this object in-place.

    Attributes:
        model (types.Model, often BaseMosaicModel): The model, typically as a subclass of :class:`BaseMosaicModel`.
        train_batch_size (int): The global batch size used for training.
        eval_batch_size (int): The batch size used for evaluation.
        grad_accum (int): The number of gradient accumulation steps to use. The size of each microbatch is ``train_batch_size / num_gpus / grad_accum``.
        max_epochs (int): The maximum number of epochs to train for.
        precision (str | Precision): The numerical precision to use for training. Should be one of ``[fp32, amp]``.
        precision_context ((precision: Precision) -> ContextManager): Function to produce a context manager to mandate precision.

        epoch (int): The index of the current epoch.
        step (int): The index of the current step/batch (measured globally).

        batch (types.Batch): The most recently retrieved batch.
        loss (types.Tensors): The most recently computed loss.
        last_batch_size (int): The size of the batch last returned from the dataloader. This can be different from the current size of ``batch`` if algorithms have modified the ``batch``.
        outputs (types.Tensors): The most recently computed output from the model's forward pass.

        optimizers (Optimizer | Tuple(Optimizer)): The optimizers being used to train the model. Multiple optimizers are not currently supported.
        schedulers (Scheduler | Tuple(Scheduler)): The learning rate schedulers, wrapped in :class:`ComposableScheduler`.
        scaler (torch.cuda.amp.GradScaler, optional): The gradient scaler in use for mixed precision training.

        train_dataloader (DataLoader): The dataloader used for training.
        eval_dataloader (DataLoader): The dataloader used for evaluation.

        algorithms (`list` of `Algorithm`): The algorithms used for training.
        callbacks (`list` of `Callback`): The callbacks used for training.
    """

    # model
    model: types.Model

    # data configurations
    train_batch_size: int
    eval_batch_size: int
    grad_accum: int

    # stopping conditions
    max_epochs: int

    # precision
    # storing precision internally so strings can be passed into the constructor and setter
    # but the getter will always return a Precision enum
    precision: Union[str, types.Precision]  # type: ignore
    _precision: types.Precision = field(init=False)  # but store an enum internally
    precision_context: Callable[[Union[str, Precision]], ContextManager] = \
        field(default_factory=default_precision_factory)

    # timing information
    epoch: int = 0  # epoch counter
    step: int = 0  # global step counter

    # transient tensors within training loop
    loss: types.Tensors = field(default_factory=make_empty_tensor)
    last_batch_size: int = 0

    batch: types.Batch = field(default_factory=dict)
    outputs: types.Tensors = field(default_factory=make_empty_tensor)

    # optimizers
    optimizers: Optional[types.Optimizers] = None
    schedulers: Optional[types.Schedulers] = None

    # scaler
    scaler: Optional[types.Scaler] = None

    # dataloaders
    train_dataloader: Optional[types.DataLoader] = None
    eval_dataloader: Optional[types.DataLoader] = None

    # algorithms
    algorithms: Sequence[Algorithm] = tuple()
    callbacks: Sequence[Callback] = tuple()

    # machine info
    world_size: int = 1
    nproc_per_node: int = 1

    # random seed
    seed: Optional[int] = None

    @property
    def global_rank(self) -> int:
        return get_global_rank()

    @property
    def local_rank(self) -> int:
        return self.global_rank % self.nproc_per_node

    @property
    def is_rank_zero(self) -> bool:
        return self.global_rank == 0

    @property
    def is_rank_set(self) -> bool:
        return is_rank_set()

    def state_dict(self) -> types.StateDict:
        """Returns the state as a :class:`dict`.
        """
        state_dict: types.StateDict = {}

        for state_field in fields(self):
            if state_field.name in SKIP_SERIALIZATION_FIELDS:
                continue
            elif state_field.name in DIRECT_SERIALIZATION_FIELDS:
                state_dict[state_field.name] = getattr(self, state_field.name)
                continue
            elif state_field.name in STATE_DICT_SERIALIZATION_FIELDS:
                state_value = getattr(self, state_field.name)
                if state_field.name == "model":
                    # Save model directly instead of by class name, since model may be wrapped by DistributedDataParallel
                    serialized_value = state_value.state_dict()
                else:
                    serialized_value = {
                        obj.__class__.__qualname__: obj.state_dict()
                        for obj in ensure_tuple(state_value)
                        if obj is not None
                    }
                state_dict[state_field.name] = serialized_value
            else:
                raise RuntimeError(f"Unable to serialize field {state_field.name}")
        state_dict["_is_model_ddp_wrapped"] = isinstance(self.model, DistributedDataParallel)
        return state_dict

    def load_state_dict(self, state: types.StateDict):
        """Loads the state.

        Args:
            state_dict (types.StateDict): object returned from call to :meth:`state_dict`.

        """
        for state_field in fields(self):
            if state_field.name in SKIP_SERIALIZATION_FIELDS:
                continue
            elif state_field.name in DIRECT_SERIALIZATION_FIELDS:
                setattr(self, state_field.name, state[state_field.name])
            elif state_field.name in STATE_DICT_SERIALIZATION_FIELDS:
                state_value = getattr(self, state_field.name)
                serialized_value = state[state_field.name]

                if state_field.name == "model":
                    if state["_is_model_ddp_wrapped"] and not isinstance(self.model, DistributedDataParallel):
                        torch.nn.modules.utils.consume_prefix_in_state_dict_if_present(serialized_value, "module.")
                    state_value.load_state_dict(serialized_value)
                else:
                    for target in ensure_tuple(state_value):
                        if target is None:
                            continue
                        if target.__class__.__qualname__ not in serialized_value:
                            warnings.warn(
                                f"{target.__class__.__qualname__} was not found in the state_dict. Its state will NOT be restored",
                                category=UserWarning)
                            continue
                        source = serialized_value[target.__class__.__qualname__]
                        target.load_state_dict(source)
            else:
                raise RuntimeError(f"Unable to load field {state_field.name}")

    @property
    def batch_idx(self) -> int:
        """int: batch_idx is the index of the batch in the current epoch."""
        return self.step - self.epoch * self.steps_per_epoch

    @property
    def steps_per_epoch(self) -> int:
        """int: The number of steps (batches) per epoch."""
        if self.train_dataloader is None:
            raise RuntimeError("To determine the number of steps per epoch, state.train_dataloader must be set.")
        return len(self.train_dataloader)

    @property
    def precision(self) -> types.Precision:
        return self._precision

    @precision.setter
    def precision(self, precision: Union[str, types.Precision]):  # type: ignore
        self._precision = Precision(precision)

    @property
    def batch_pair(self) -> types.BatchPair:
        """:obj:`BatchPair`: The current batch, represented as a `BatchPair`.

        :raises TypeError: If the current batch is not a `BatchPair`.
        """
        return types.as_batch_pair(self.batch)

    @property
    def batch_dict(self) -> types.BatchDict:
        """:obj:`BatchDict`: The current batch, represented as a `BatchDict`.

        :raises TypeError: If the current batch is not a `BatchDict`.
        """
        return types.as_batch_dict(self.batch)



# Copyright 2021 MosaicML. All Rights Reserved.

import pathlib
import random
from dataclasses import Field, fields

import pytest
import torch
import torch.distributed as dist
import torch.nn.functional as F
from _pytest.monkeypatch import MonkeyPatch
from torch.functional import Tensor

from composer.algorithms.dummy import DummyHparams
from composer.core import State, types
from composer.core.state import DIRECT_SERIALIZATION_FIELDS, SKIP_SERIALIZATION_FIELDS, STATE_DICT_SERIALIZATION_FIELDS
from composer.models.resnet56_cifar10.resnet56_cifar10_hparams import CIFARResNetHparams
from composer.utils import ensure_tuple


def random_tensor(size=(4, 10)):
    return torch.rand(*size)


class CIFARResNet():
    """
    CIFAR ResNet model with corresponding random data generation
    """

    @staticmethod
    def get_model():
        return CIFARResNetHparams(num_classes=10).initialize_object()

    @staticmethod
    def random_data_pair(N=8, C=3, H=32, W=32, n_classes=10):
        input = torch.rand(N, C, H, W)
        label = torch.randint(0, n_classes, size=(N,))
        return input, label


def get_dummy_state():
    # TODO: use a smaller test model
    dummy_model = CIFARResNet.get_model()
    optimizers = torch.optim.Adadelta(dummy_model.parameters())

    return State(model=dummy_model,
                 train_batch_size=random.randint(0, 100),
                 eval_batch_size=random.randint(0, 100),
                 grad_accum=random.randint(0, 100),
                 precision=types.Precision.AMP,
                 max_epochs=random.randint(0, 100),
                 epoch=random.randint(0, 100),
                 step=random.randint(0, 100),
                 loss=random_tensor(),
                 batch=(random_tensor(), random_tensor()),
                 outputs=random_tensor(),
                 optimizers=optimizers,
                 schedulers=torch.optim.lr_scheduler.StepLR(optimizers, step_size=3),
                 algorithms=[DummyHparams().initialize_object()])


def is_field_serialized(f: Field) -> bool:
    if f.name in STATE_DICT_SERIALIZATION_FIELDS or f.name in DIRECT_SERIALIZATION_FIELDS:
        return True
    elif f.name in SKIP_SERIALIZATION_FIELDS:
        return False
    else:
        raise RuntimeError(f"Serialization method for field {f.name} not specified")


def assert_state_equivalent(state1: State, state2: State, skip_transient_fields: bool):
    # tested separately
    IGNORE_FIELDS = [
        'optimizers',
        'schedulers',
        'train_dataloader',
        'eval_dataloader',
        'algorithms',
        'callbacks',
        'precision_context',
    ]

    for f in fields(state1):
        if f.name in IGNORE_FIELDS:
            continue
        if skip_transient_fields and not is_field_serialized(f):
            continue

        var1 = getattr(state1, f.name)
        var2 = getattr(state2, f.name)

        if f.name == "model":
            for p, q in zip(state1.model.parameters(), state2.model.parameters()):
                torch.testing.assert_allclose(p, q, atol=1e-2, rtol=1e-2)
        elif isinstance(var1, types.Tensor):
            torch.testing.assert_equal(var1, var2)
        else:
            assert var1 == var2


def train_one_step(state: State, batch: types.Batch) -> None:
    x, y = batch
    state.batch = batch

    state.outputs = state.model(state.batch_pair)
    assert isinstance(y, Tensor)

    state.loss = F.cross_entropy(state.outputs, y)
    state.loss.backward()
    assert state.optimizers is not None
    assert state.schedulers is not None
    for optimizer in ensure_tuple(state.optimizers):
        optimizer.step()
    for scheduler in ensure_tuple(state.schedulers):
        scheduler.step()
    state.step += 1


@pytest.mark.run_long
@pytest.mark.timeout(5)
def test_state_serialize(tmpdir: pathlib.Path):

    state1 = get_dummy_state()
    state2 = get_dummy_state()

    # train one step to set the optimizer states
    batch = CIFARResNet.random_data_pair()
    train_one_step(state1, batch)

    # load from state1 to state2
    state_dict = state1.state_dict()
    filepath = str(tmpdir / "state.pt")
    torch.save(state_dict, filepath)
    state_dict_2 = torch.load(filepath, map_location="cpu")
    state2.load_state_dict(state_dict_2)
    # Make sure there was nothing wrong serialization/deserialization of permanent
    assert_state_equivalent(state1, state2, skip_transient_fields=True)

    # train both for one step on another sample
    batch = CIFARResNet.random_data_pair()
    train_one_step(state1, batch)
    train_one_step(state2, batch)

    # both states should have equivalent state, model parameters, loss, and outputs
    assert_state_equivalent(state1, state2, skip_transient_fields=False)


def test_state_rank(monkeypatch: MonkeyPatch, dummy_state: State):
    monkeypatch.setattr(dist, "get_rank", lambda: 9)
    dummy_state.nproc_per_node = 4
    assert dummy_state.is_rank_set
    assert not dummy_state.is_rank_zero
    assert dummy_state.local_rank == 1
    assert dummy_state.global_rank == 9



# Copyright 2021 MosaicML. All Rights Reserved.

from __future__ import annotations

from contextlib import contextmanager
from typing import Generator, Optional, Union

import torch
import torch.cuda.amp
import torch.utils.data

from composer.core.state import State
from composer.core.types import Batch, DataLoader, Precision, StateDict, Tensor, TPrefetchFn
from composer.datasets.dataloader import WrappedDataLoader
from composer.trainer.devices.device import Device, T_nnModule
from composer.utils import map_collection


class CudaDataLoader(WrappedDataLoader):
    """
    CudaDataLoader wraps a DataLoader and moves samples onto the specified device
    as they are used
    """

    def __init__(
        self,
        dataloader: DataLoader,
        prefetch_in_cuda_stream: bool,
        device: torch.device,
        prefetch_fn: Optional[TPrefetchFn] = None,
    ) -> None:
        """
        dataloader: The dataloader to wrap
        device: The device that samples should automatically be moved to upon iteration
        """
        super().__init__(dataloader)
        self.device = device
        self.prefetch_in_cuda_stream = prefetch_in_cuda_stream
        self.prefetch_fn = prefetch_fn

    def __iter__(self):
        if self.prefetch_in_cuda_stream:
            return self.prefetch_in_cuda_stream_iter()
        else:
            return self.normal_iter()

    def normal_iter(self):
        for batch in self.dataloader:
            batch = self.move_to_gpu(batch)
            if self.prefetch_fn is not None:
                batch = self.prefetch_fn(batch)
            yield batch

    def prefetch_in_cuda_stream_iter(self):
        stream = torch.cuda.Stream()
        batch: Optional[Batch] = None

        for next_batch in self.dataloader:
            with torch.cuda.stream(stream):
                next_batch = self.move_to_gpu(next_batch)
                if self.prefetch_fn is not None:
                    next_batch = self.prefetch_fn(next_batch)
            if batch is not None:
                yield batch

            torch.cuda.current_stream().wait_stream(stream)
            batch = next_batch
        if batch is not None:
            yield batch

    def _to_device(self, x: Tensor) -> Tensor:
        return x.to(self.device, non_blocking=True)

    def move_to_gpu(self, batch: Batch) -> Batch:
        if isinstance(batch, Tensor):
            return self._to_device(batch)
        if isinstance(batch, (tuple, list)):
            return type(batch)(self._to_device(x) for x in batch)
        if isinstance(batch, dict):
            return {k: self._to_device(v) for k, v in batch.items()}

        return map_collection(batch, self._to_device)  # type: ignore


class DeviceGPU(Device):

    def __init__(
        self,
        prefetch_in_cuda_stream: bool,
        n_gpus: int,
    ):
        self.n_gpus = n_gpus
        self.prefetch_in_cuda_stream = prefetch_in_cuda_stream
        self._device: Optional[torch.device] = None

    @property
    def nproc_per_node(self) -> int:
        return self.n_gpus

    def prepare(self, state: State) -> None:
        if self._device is not None:
            raise ValueError("device is already set")
        gpu = state.local_rank
        self._device = torch.device(f"cuda:{gpu}")
        torch.cuda.set_device(self._device)
        assert torch.cuda.current_device() == gpu

    def _require_device(self) -> None:
        if self._device is None:
            raise RuntimeError("device.module_to_device() was called before device.prepare(). "
                               "Please call device.prepare() first.")

    def module_to_device(self, module: T_nnModule) -> T_nnModule:
        self._require_device()
        assert self._device is not None
        return module.to(self._device)

    def dataloader_to_device(self, dataloader: DataLoader, prefetch_fn: Optional[TPrefetchFn]) -> DataLoader:
        self._require_device()
        assert self._device is not None
        return CudaDataLoader(
            dataloader,
            device=self._device,
            prefetch_in_cuda_stream=self.prefetch_in_cuda_stream,
            prefetch_fn=prefetch_fn,
        )

    def tensor_to_device(self, tensor: Tensor) -> Tensor:
        return tensor.to(self._device, non_blocking=True)

    @contextmanager
    def precision_context(self, precision: Union[str, Precision]) -> Generator[None, None, None]:
        precision = Precision(precision)
        if precision == Precision.FP32:
            enabled = False
        elif precision == Precision.AMP:
            enabled = True
        else:
            raise ValueError(f"Precision {precision} not supported for a GPU")
        with torch.cuda.amp.autocast(enabled):  #type: ignore
            yield

    @property
    def ddp_backend(self) -> str:
        return "nccl"

    def state_dict(self) -> StateDict:
        return {
            "rng": torch.cuda.get_rng_state(),
        }

    def load_state_dict(self, state: StateDict) -> None:
        torch.cuda.set_rng_state(state["rng"])



# Copyright 2021 MosaicML. All Rights Reserved.

import itertools

import pytest
import torch

from composer.algorithms import blurpool
from composer.algorithms.blurpool.blurpool import BlurPoolHparams
from composer.trainer import TrainerHparams
from tests.utils.trainer_fit import train_model


def generate_pool_args():
    n_vals = [2]
    c_vals = [2]
    size_vals = [(3, 3), (3, 7), (4, 4)]
    strides = [1, 2]
    filter_size_vals = [(1, 1), (1, 3), (3, 3)]
    return list(itertools.product(n_vals, c_vals, size_vals, strides, filter_size_vals))


@pytest.mark.parametrize('pool_args', generate_pool_args())
def test_blurmaxpool_shapes(pool_args):
    n, c, sz, stride, kernel_size = pool_args

    X = torch.randn(n, c, sz[0], sz[1])

    layer_args = dict(kernel_size=kernel_size, stride=stride, dilation=1)
    blurpool_layer = blurpool.BlurMaxPool2d(**layer_args)
    maxpool_layer = torch.nn.MaxPool2d(**layer_args)

    assert blurpool_layer(X).shape == maxpool_layer(X).shape


@pytest.mark.parametrize('blur_first', [True, False])
@pytest.mark.parametrize('pool_args', generate_pool_args())
def test_blurconv2d_shapes(pool_args, blur_first):
    n, c, sz, stride, kernel_size = pool_args

    X = torch.randn(n, c, sz[0], sz[1])

    layer_args = dict(kernel_size=kernel_size, stride=stride, dilation=1, in_channels=c, out_channels=(c + 1))
    blurconv2d_layer = blurpool.BlurConv2d(**layer_args, blur_first=blur_first)
    conv2d_layer = torch.nn.Conv2d(**layer_args)

    assert blurconv2d_layer(X).shape == conv2d_layer(X).shape


@pytest.mark.parametrize('pool_args', generate_pool_args())
def test_blur2d_shapes(pool_args):
    n, c, sz, _, _ = pool_args

    X = torch.randn(n, c, sz[0], sz[1])
    out = blurpool.blur_2d(X)
    assert out.shape == X.shape


def test_default_2d_filter():

    def reference_filter():
        filt = torch.FloatTensor([1, 2, 1])
        filt = torch.outer(filt, filt)
        filt *= 1. / filt.sum()
        filt = torch.Tensor(filt)
        return filt.view(1, 1, *filt.shape)

    torch.testing.assert_allclose(
        blurpool.blurpool_layers._default_2d_filter(),  # type: ignore
        reference_filter(),
    )


@pytest.mark.parametrize('pool_args', generate_pool_args())
def test_blur2d_std(pool_args):
    n, c, sz, _, _ = pool_args

    X = torch.randn(n, c, sz[0], sz[1])
    out = blurpool.blur_2d(X)
    assert torch.std(out) <= torch.std(X)


def test_blurpool_blurconv2d_params_match_original_params():
    conv2d = torch.nn.Conv2d(16, 32, 3, stride=1, bias=True)
    blurconv = blurpool.BlurConv2d.from_conv2d(conv2d)
    torch.testing.assert_allclose(blurconv.conv.weight, conv2d.weight)
    torch.testing.assert_allclose(blurconv.conv.bias, conv2d.bias)
    assert blurconv.conv.weight.requires_grad
    assert blurconv.conv.bias is not None
    assert blurconv.conv.bias.requires_grad


@pytest.mark.run_long
@pytest.mark.timeout(90)
def test_blurpool_trains(mosaic_trainer_hparams: TrainerHparams):
    mosaic_trainer_hparams.algorithms = [BlurPoolHparams(replace_convs=True, replace_maxpools=True, blur_first=True)]
    train_model(mosaic_trainer_hparams, run_loss_check=True)



# Copyright 2021 MosaicML. All Rights Reserved.

from __future__ import annotations

import time
from collections import deque
from dataclasses import dataclass
from typing import Deque, Optional

import yahp as hp

from composer import Logger, State
from composer.callbacks.callback_hparams import CallbackHparams
from composer.core.callback import RankZeroCallback
from composer.core.types import StateDict


class SpeedMonitor(RankZeroCallback):
    """Callback to monitor the training throughput.

    It logs:

    * A rolling average (over the :attr:`window_size` most recent batches)
      of the number of samples processed per second to the
      ``throughput/step`` key.
    * The number of samples processed per second, averaged over
      an entire epoch, to the ``throughput/epoch`` key.
    * The total elapsed training time to the ``wall_clock_train`` key.

    Args:
        window_size (int):
            Number of batchs to use for a rolling average of throughput.
    """

    def __init__(self, window_size: int):
        super().__init__()
        self.train_examples_per_epoch = 0
        self.wall_clock_train = 0.0
        self.epoch_start_time = 0.0
        self.batch_end_times: Deque[float] = deque(maxlen=window_size + 1)  # rolling list of batch end times
        self.batch_num_samples: Deque[int] = deque(maxlen=window_size)  # rolling list of num samples in batch.
        self.hparams = SpeedMonitorHparams(window_size=window_size)
        self.loaded_state: Optional[StateDict] = None

    def state_dict(self) -> StateDict:
        current_time = time.time()
        return {
            "train_examples_per_epoch": self.train_examples_per_epoch,
            "wall_clock_train": self.wall_clock_train,
            "epoch_duration": current_time - self.epoch_start_time,
            "batch_durations": [current_time - x for x in self.batch_end_times],
            "batch_num_samples": self.batch_num_samples,
        }

    def load_state_dict(self, state: StateDict) -> None:
        self.loaded_state = state

    def _load_state(self) -> None:
        current_time = time.time()
        if self.loaded_state is not None:
            self.train_examples_per_epoch = self.loaded_state["train_examples_per_epoch"]
            self.wall_clock_train = self.loaded_state["wall_clock_train"]
            self.epoch_start_time = current_time - self.loaded_state["epoch_duration"]
            self.batch_end_times = deque([current_time - x for x in self.loaded_state["batch_durations"]],
                                         maxlen=self.hparams.window_size + 1)
            self.batch_num_samples = self.loaded_state["batch_num_samples"]
            self.loaded_state = None

    def batch_start(self, state: State, logger: Logger) -> None:
        self._load_state()

    def epoch_start(self, state: State, logger: Logger):
        self._load_state()
        self.epoch_start_time = time.time()
        self.batch_end_times.clear()
        self.batch_num_samples.clear()
        self.train_examples_per_epoch = 0

    def batch_end(self, state: State, logger: Logger):
        self.batch_end_times.append(time.time())
        batch_num_samples = 0
        batch_num_samples += state.last_batch_size
        # TODO this is a hack around not having proper syncing / reduction available in callbacks
        # Ideally, callbacks would have a way of reducing tensors.
        # It assumes that each process has equal batch sizing
        # For the speed monitor, we might be able to use the static step converter with num_samples
        batch_num_samples *= state.world_size
        self.batch_num_samples.append(batch_num_samples)
        self.train_examples_per_epoch += batch_num_samples
        if len(self.batch_end_times) == self.hparams.window_size + 1:
            throughput = sum(self.batch_num_samples) / (self.batch_end_times[-1] - self.batch_end_times[0])
            logger.metric_batch({'throughput/step': throughput})

    def epoch_end(self, state: State, logger: Logger):
        epoch_time = time.time() - self.epoch_start_time
        self.wall_clock_train += epoch_time
        logger.metric_epoch({
            "wall_clock_train": self.wall_clock_train,
        })
        logger.metric_epoch({
            "throughput/epoch": self.train_examples_per_epoch / epoch_time,
        })


@dataclass
class SpeedMonitorHparams(CallbackHparams):
    """Parameters for the :class:`SpeedMonitor`.

    See the documentation for the :class:`SpeedMonitor`.
    """
    window_size: int = hp.optional(
        doc="Number of batchs to use for a rolling average of throughput.",
        default=100,
    )

    def initialize_object(self) -> SpeedMonitor:
        return SpeedMonitor(window_size=self.window_size)



# Copyright 2021 MosaicML. All Rights Reserved.

from typing import Dict, Type

import pytest
import torch

from composer.core.types import ModelParameters
from composer.optim.optimizer_hparams import (AdamHparams, AdamWHparams, DecoupledAdamWHparams, DecoupledSGDWHparams,
                                              OptimizerHparams, RAdamHparams, RMSPropHparams, SGDHparams, get_optimizer)
from composer.trainer.trainer_hparams import optimizer_registry

optimizer_constructors: Dict[Type[OptimizerHparams], OptimizerHparams] = {
    AdamHparams: AdamHparams(),
    RAdamHparams: RAdamHparams(),
    AdamWHparams: AdamWHparams(),
    DecoupledAdamWHparams: DecoupledAdamWHparams(),
    SGDHparams: SGDHparams(lr=0.001),
    DecoupledSGDWHparams: DecoupledSGDWHparams(lr=0.001),
    RMSPropHparams: RMSPropHparams(lr=0.001),
}


@pytest.fixture
def dummy_parameters() -> ModelParameters:
    sizes = [(5, 5), (10, 10)]
    parameters = [torch.rand(size) for size in sizes]
    return parameters


@pytest.mark.parametrize("optimizer_name", optimizer_registry.keys())
def test_optimizer_initialization(optimizer_name, dummy_parameters):

    # create the optimizer hparams object
    optimizer_cls: Type[OptimizerHparams] = optimizer_registry[optimizer_name]
    optimizer_hparams = optimizer_constructors[optimizer_cls]

    # create the optimizer object using the hparams
    optimizer = get_optimizer(dummy_parameters, optimizer_hparams)
    assert isinstance(optimizer, optimizer_hparams.optimizer_object)



# Copyright 2021 MosaicML. All Rights Reserved.

from composer.algorithms.progressive_resizing.progressive_resizing import ProgressiveResizing as ProgressiveResizing
from composer.algorithms.progressive_resizing.progressive_resizing import \
    ProgressiveResizingHparams as ProgressiveResizingHparams
from composer.algorithms.progressive_resizing.progressive_resizing import resize_inputs as resize_inputs

_name = 'Progressive Resizing'
_class_name = 'ProgressiveResizing'
_functional = 'resize_inputs'
_tldr = 'Increases the input image size during training'
_attribution = 'Fast AI'
_link = 'https://github.com/fastai/fastbook/blob/780b76bef3127ce5b64f8230fce60e915a7e0735/07_sizing_and_tta.ipynb'
_method_card = ''



# Copyright 2021 MosaicML. All Rights Reserved.

import pytest

from composer.core import Event


@pytest.mark.parametrize('event', list(Event))
def test_event_values(event: Event):
    assert event.name.lower() == event.value



# Copyright 2021 MosaicML. All Rights Reserved.

from typing import Dict, List, Type

from composer.algorithms.algorithm_hparams import AlgorithmHparams
from composer.algorithms.alibi import AlibiHparams
from composer.algorithms.augmix import AugMixHparams
from composer.algorithms.blurpool import BlurPoolHparams
from composer.algorithms.channels_last import ChannelsLastHparams
from composer.algorithms.colout import ColOutHparams
from composer.algorithms.curriculum_learning import CurriculumLearningHparams
from composer.algorithms.cutout import CutOutHparams
from composer.algorithms.dummy import DummyHparams
from composer.algorithms.ghost_batchnorm import GhostBatchNormHparams
from composer.algorithms.label_smoothing import LabelSmoothingHparams
from composer.algorithms.layer_freezing import LayerFreezingHparams
from composer.algorithms.mixup import MixUpHparams
from composer.algorithms.no_op_model import NoOpModelHparams
from composer.algorithms.progressive_resizing import ProgressiveResizingHparams
from composer.algorithms.randaugment import RandAugmentHparams
from composer.algorithms.sam import SAMHparams
from composer.algorithms.scale_schedule import ScaleScheduleHparams
from composer.algorithms.selective_backprop import SelectiveBackpropHparams
from composer.algorithms.squeeze_excite import SqueezeExciteHparams
from composer.algorithms.stochastic_depth import StochasticDepthHparams
from composer.algorithms.swa.hparams import SWAHparams
from composer.core.algorithm import Algorithm

registry: Dict[str, Type[AlgorithmHparams]] = {
    'dummy': DummyHparams,
    'blurpool': BlurPoolHparams,
    'channels_last': ChannelsLastHparams,
    'curriculum_learning': CurriculumLearningHparams,
    'cutout': CutOutHparams,
    'ghost_batchnorm': GhostBatchNormHparams,
    'label_smoothing': LabelSmoothingHparams,
    'layer_freezing': LayerFreezingHparams,
    'squeeze_excite': SqueezeExciteHparams,
    'swa': SWAHparams,
    'no_op_model': NoOpModelHparams,
    'mixup': MixUpHparams,
    'no_op_model': NoOpModelHparams,
    'scale_schedule': ScaleScheduleHparams,
    'stochastic_depth': StochasticDepthHparams,
    'colout': ColOutHparams,
    'progressive_resizing': ProgressiveResizingHparams,
    'randaugment': RandAugmentHparams,
    'augmix': AugMixHparams,
    'sam': SAMHparams,
    'alibi': AlibiHparams,
    'selective_backprop': SelectiveBackpropHparams,
}


def get_algorithm_registry():
    return registry


def get_algorithm(params: AlgorithmHparams) -> Algorithm:
    return params.initialize_object()


def list_algorithms() -> List[str]:
    return list(registry.keys())



# Copyright 2021 MosaicML. All Rights Reserved.

from __future__ import annotations

import logging
from dataclasses import asdict, dataclass
from typing import Optional

import torch
import yahp as hp

from composer.algorithms.algorithm_hparams import AlgorithmHparams
from composer.core import Algorithm, Event, Logger, State, surgery

log = logging.getLogger(__name__)


@dataclass
class SqueezeExciteHparams(AlgorithmHparams):

    latent_channels: float = hp.optional(
        doc='Dimensionality of hidden layer within the added MLP.',
        default=64,
    )
    min_channels: int = hp.optional(
        doc='Minimum number of channels in a Conv2d layer'
        ' for a squeeze-excite block to be placed after it.',
        default=128,
    )

    def initialize_object(self) -> SqueezeExcite:
        return SqueezeExcite(**asdict(self))


class SqueezeExcite2d(torch.nn.Module):
    """`Squeeze-and-Excitation block <https://arxiv.org/abs/1709.01507>`_"""

    def __init__(self, num_features, latent_channels=.125):
        super().__init__()
        self.latent_channels = int(latent_channels if latent_channels >= 1 else latent_channels * num_features)
        flattened_dims = num_features

        self.pool_and_mlp = torch.nn.Sequential(torch.nn.AdaptiveAvgPool2d(1), torch.nn.Flatten(),
                                                torch.nn.Linear(flattened_dims, self.latent_channels, bias=False),
                                                torch.nn.ReLU(),
                                                torch.nn.Linear(self.latent_channels, num_features, bias=False),
                                                torch.nn.Sigmoid())

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        n, c, _, _ = input.shape
        attention_coeffs = self.pool_and_mlp(input)
        return input * attention_coeffs.reshape(n, c, 1, 1)


class SqueezeExciteConv2d(torch.nn.Module):
    """Helper class used to add a `Squeeze-and-Excitation block <https://arxiv.org/abs/1709.01507>`_ after a Conv2d."""

    def __init__(self, *args, latent_channels=.125, conv: torch.nn.Conv2d = None, **kwargs):
        super().__init__()
        self.conv = torch.nn.Conv2d(*args, **kwargs) if conv is None else conv
        self.se = SqueezeExcite2d(num_features=self.conv.out_channels, latent_channels=latent_channels)

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        return self.se(self.conv(input))

    @staticmethod
    def from_conv2d(module: torch.nn.Conv2d, module_index: int, latent_channels: float):
        return SqueezeExciteConv2d(conv=module, latent_channels=latent_channels)


def apply_se(model: torch.nn.Module, latent_channels: float, min_channels: int):
    """Adds Squeeze-and-Excitation <https://arxiv.org/abs/1709.01507>`_  (SE) blocks
    after the `Conv2d` layers of a neural network.

    Args:
        model: A module containing one or more `torch.nn.Conv2d` modules.
        latent_channels: The dimensionality of the hidden layer within the added MLP.
        min_channels: An SE block is added after a `Conv2d` module `conv`
            only if `min(conv.in_channels, conv.out_channels) >= min_channels`.
            For models that reduce spatial size and increase channel count
            deeper in the network, this parameter can be used to only
            add SE blocks deeper in the network. This may be desirable
            because SE blocks add less overhead when their inputs have
            smaller spatial size.
    """

    def convert_module(module: torch.nn.Conv2d, module_index: int):
        if min(module.in_channels, module.out_channels) < min_channels:
            return None
        return SqueezeExciteConv2d.from_conv2d(module, module_index, latent_channels=latent_channels)

    transforms = {torch.nn.Conv2d: convert_module}
    surgery.replace_module_classes(model, transforms)  # type: ignore
    return model


class SqueezeExcite(Algorithm):
    """Adds Squeeze-and-Excitation <https://arxiv.org/abs/1709.01507>`_  (SE) blocks
    after the `Conv2d` layers of a neural network.

    Args:
        latent_channels: The dimensionality of the hidden layer within the added MLP.
        min_channels: An SE block is added after a `Conv2d` module `conv`
            only if `min(conv.in_channels, conv.out_channels) >= min_channels`.
            For models that reduce spatial size and increase channel count
            deeper in the network, this parameter can be used to only
            add SE blocks deeper in the network. This may be desirable
            because SE blocks add less overhead when their inputs have
            smaller spatial size.
    """

    def __init__(
        self,
        latent_channels: float = 64,
        min_channels: int = 128,
    ):
        """
        __init__ is constructed from the same fields as in hparams.
        """
        self.hparams = SqueezeExciteHparams(
            latent_channels=latent_channels,
            min_channels=min_channels,
        )

    def match(self, event: Event, state: State) -> bool:
        return event == Event.INIT

    def apply(self, event: Event, state: State, logger: Logger) -> Optional[int]:
        """
        Apply the Squeeze-and-Excitation layer replacement.
        """
        state.model = apply_se(state.model,
                               latent_channels=self.hparams.latent_channels,
                               min_channels=self.hparams.min_channels)
        layer_count = surgery.count_module_instances(state.model, SqueezeExciteConv2d)

        log.info(f'Applied SqueezeExcite to model {state.model.__class__.__name__} '
                 f'with latent_channels={self.hparams.latent_channels}, '
                 f'min_channels={self.hparams.min_channels}. '
                 f'Model now has {layer_count} SqueezeExcite layers.')

        logger.metric_fit({
            'squeeze_excite/num_squeeze_excite_layers': layer_count,
        })



# Copyright 2021 MosaicML. All Rights Reserved.

import logging
from collections import OrderedDict
from dataclasses import dataclass
from typing import Dict, Optional, Sequence, Union

from composer.core.algorithm import Algorithm
from composer.core.callback import Callback
from composer.core.event import Event
from composer.core.logging import Logger
from composer.core.state import State

log = logging.getLogger(__name__)
Traces = Dict[str, "Trace"]


@dataclass
class Trace():
    """
    Record of an algorithm's execution.

    Attributes:
        exit_code (int or None): optional return value from algorithm. Default: None.
        order (int or None): order in which the algorithm was executed
                             in the list of algorithms. None means algorithm was not run.
        run (bool): whether the algorithm was run. Default: False
    """
    exit_code: Optional[int] = None
    order: Optional[int] = None
    run: bool = False


def _setup_trace(algorithms: Sequence[Algorithm], event: Event) -> Traces:
    """
    The default traces of an entire run is an OrderedDict, with the keys
    of format 'algorithm_name/event' (e.g. Blurpool/TRAINING_START).
    """
    return OrderedDict([(f'{algo}/{event}', Trace()) for algo in algorithms])


class Engine():
    """
    Coordinator for running and resolving conflicts between algorithms.

    Args:
        state (State): the initial ``State`` of the trainer. Will be modified in-place.
        algorithms (Sequence[Algorithm]): the list of algorithms for this engine to execute.
        logger (Optional[Logger]): a ``Logger`` instance to be used for logging algorithm and callback specific metrics.
        callbacks (Sequence[Callback]): the list of callbacks for this engine to execute.
    """

    def __init__(self,
                 state: State,
                 algorithms: Sequence[Algorithm],
                 logger: Optional[Logger] = None,
                 callbacks: Sequence[Callback] = None):
        if logger is None:
            log.warning("No logger passed to the engine.  Defaulting to an empty logger")
            logger = Logger(state=state, log_destinations=[])

        assert logger is not None
        self.logger = logger
        self.state = state
        self.algorithms = algorithms
        self.callbacks = callbacks or []

    def run_event(
        self,
        event: Union[Event, str],
    ) -> Traces:
        """Runs the sequence of algorithms and callbacks.
        
        Filters algorithms by calling each one's ``match`` function, then passes to ``compile`` to
        make any modifications, then runs each algorithm's ``apply`` function to make in-place
        changes to the ``State``.

        The order of algorithm execution is determined by the provided list, and any changes made
        by ``compile``.

        Returns ``Traces`` of the execution, a dictionary with keys formatted as 'algorithm_name/event'
        (e.g. Blurpool/TRAINING_START), and values are the ``Trace`` object, which include an optional
        return code from the algorithm, the order of execution, and whether the algorithm was run.

        Callbacks are always ran after algorithms, and do not return a trace.

        Can be called with either the Event enum, or a string of the event value.

        Examples:
            >>> engine = Engine(state, algorithms, logger, callbacks)
            >>> engine.run_event(Event.BEFORE_LOSS) # or
            >>> engine.run_event('before_loss') # also works


        Args:
            event (Event or str): the current ``Event``. Can be the Enum or a string with the event value.
        Returns:
            traces (Dict[str, Trace]): dictionary of trace for each algorithm.
        """
        traces = self._run_algorithms(event)
        self._run_callbacks(event)
        return traces

    def _run_algorithms(
        self,
        event: Union[Event, str],
    ) -> Traces:
        event = Event(event)

        algorithms_to_run = [algo for algo in self.algorithms if algo.match(event, self.state)]

        # future collision resolution
        algorithms_to_run = self._compile(algorithms_to_run, event)

        trace = _setup_trace(algorithms_to_run, event)
        for order, algorithm in enumerate(algorithms_to_run):
            exit_code = algorithm.apply(event, self.state, self.logger)

            trace_key = f'{algorithm}/{event}'
            trace[trace_key] = Trace(exit_code=exit_code, order=order, run=True)

        if self.logger is not None:
            self.logger.metric_verbose(data={key: 1 if tr.run else 0 for key, tr in trace.items()})

        return trace

    def _compile(
        self,
        algorithms_to_run: Sequence[Algorithm],
        event: Union[Event, str],
    ) -> Sequence[Algorithm]:
        """
        Runs compiliation passes that modify the order and content of a list of algorithms.

        Currently, runs the algorithms in a FILO queue for the before_ and after_ events. For example,
        algorithms will run in order ABCD during before_loss, and in DCBA during after_loss. The motivation
        here is that algorithms can 'undo' their effects upon the exit of an event. Note that events that
        have the pattern _start or _end will still run with ABCD order.

        Intent of this method is to eventually store and handle other algorithms collisions and ordering
        requirements.

        Args:
            algorithms_to_run(Sequence[Algorithm]): sequence of algorithms
            event (Event): the current event

        Returns:
            algorithms_to_run(Sequence[Algorithm]): modified sequence of algorithms
        """
        event = Event(event)

        if event.value.startswith('after') or event.value.startswith('eval_after'):
            """
            Establish a FILO queue of algorithms before_ and after_ an event.
            before_loss: A, B, C, D
            after_loss: D, C, B, A

            TODO: store an actual LIFO queue upon __enter__ and __exit__
            """
            algorithms_to_run = list(reversed(algorithms_to_run))

        return algorithms_to_run

    def _run_callbacks(
        self,
        event: Union[Event, str],
    ):
        """
        Runs a sequence of callbacks, by calling the method corresponding to the event.
        Args:
            event (Event): the current ``Event``
            state (State): the current ``State``
            callbacks (Sequence[Callback]): a sequence of callbacks
        Returns:
            None
        """
        event = Event(event)

        for cb in self.callbacks:
            if not hasattr(cb, event.value):
                raise ValueError(f'f{cb} has no method for event {event}')
            else:
                f = getattr(cb, event.value)
                f(self.state, self.logger)



# Copyright 2021 MosaicML. All Rights Reserved.

from composer.datasets.brats import BratsDatasetHparams as BratsDatasetHparams
from composer.datasets.cifar10 import CIFAR10DatasetHparams as CIFAR10DatasetHparams
from composer.datasets.dataloader import DataloaderHparams as DataloaderHparams
from composer.datasets.dataloader import WrappedDataLoader as WrappedDataLoader
from composer.datasets.hparams import DataloaderSpec as DataloaderSpec
from composer.datasets.hparams import DatasetHparams as DatasetHparams
from composer.datasets.imagenet import ImagenetDatasetHparams as ImagenetDatasetHparams
from composer.datasets.lm_datasets import LMDatasetHparams as LMDatasetHparams
from composer.datasets.mnist import MNISTDatasetHparams as MNISTDatasetHparams
from composer.datasets.synthetic import MemoryFormat as MemoryFormat
from composer.datasets.synthetic import SyntheticDataset as SyntheticDataset
from composer.datasets.synthetic import SyntheticDatasetHparams as SyntheticDatasetHparams



# Copyright 2021 MosaicML. All Rights Reserved.

from dataclasses import dataclass

import yahp as hp

from composer.callbacks.callback_hparams import CallbackHparams
from composer.core import Logger, State
from composer.core.callback import Callback


class GradMonitor(Callback):
    """Callback to log the L2 norm of all the parameters in the model,
    and optionally, the L2 norm of the parameters by each layer.

    It logs the L2 norm on each batch under the ``grad_l2_norm/step`` key.
    If :attr:`log_layer_grad_norms` is True (default False), then
    layer-wise L2 norms are logged under ``layer_grad_l2_norm/LAYER_NAME``.

    Args:
        log_layer_grad_norms (bool, optional):
            Whether to log the L2 normalization of each layer.
            Defaults to False.
    """

    def __init__(self, log_layer_grad_norms: bool = False):
        super().__init__()
        self.log_layer_grad_norms = log_layer_grad_norms

    def after_train_batch(self, state: State, logger: Logger):
        """Compute the gradient L2 norm after the reduction of the
        backwards pass across GPUs. This function iterates over the
        parameters of the model and hence may cause a reduction in
        throughput while training large models. In order to ensure
        correctness, this function should be called after gradient
        unscaling in cases where gradients are scaled.

        Args:
            state (State): The :class:`~composer.core.State` object
                used during training.
            logger (Logger):
                The :class:`~composer.core.logging.logger.Logger` object.
        """
        norm = None
        layer_norms = {}
        for name, p in state.model.named_parameters():
            if p.grad is not None and p.requires_grad:
                param_grad_norm = p.grad.detach().data.norm(2).item()  # type: ignore
                if self.log_layer_grad_norms:
                    layer_norms[f'layer_grad_l2_norm/{name}'] = param_grad_norm

                param_grad_norm = param_grad_norm**2
                norm = param_grad_norm if not norm else norm + param_grad_norm

        norm = norm**0.5
        logger.metric_batch({'grad_l2_norm/step': norm})
        if self.log_layer_grad_norms:
            logger.metric_batch(layer_norms)


@dataclass
class GradMonitorHparams(CallbackHparams):
    """Parameters for the :class:`GradMonitor`.

    See the documentation for the :class:`GradMonitor`.
    """

    log_layer_grad_norms: bool = hp.optional(
        doc="Whether to log gradient norms for individual layers.",
        default=False,
    )

    def initialize_object(self) -> GradMonitor:
        return GradMonitor(log_layer_grad_norms=self.log_layer_grad_norms)



# Copyright 2021 MosaicML. All Rights Reserved.

from composer.models.resnet50.model import ResNet50 as ResNet50
from composer.models.resnet50.resnet50_hparams import ResNet50Hparams as ResNet50Hparams

_task = 'Image Classification'
_dataset = 'ImageNet'
_name = 'ResNet50'
_quality = 76.5
_metric = 'Top-1 Accuracy'
_ttt = '?'
_hparams = 'hparams.yaml'



# Copyright 2021 MosaicML. All Rights Reserved.

from __future__ import annotations

import importlib
import logging
import math
from dataclasses import asdict, dataclass
from operator import attrgetter
from types import MethodType, ModuleType
from typing import Any, Callable, Optional, Union

import torch
import yahp as hp

from composer.algorithms import AlgorithmHparams
from composer.core import Algorithm, Event, Logger, State, surgery

log = logging.getLogger(__name__)


@dataclass
class AlibiHparams(AlgorithmHparams):

    position_embedding_attribute: str = hp.required("attribute for position embeddings. "
                                                    "For example in HuggingFace's GPT2, the position "
                                                    "embeddings are 'transformer.wpe'")
    attention_module_name: str = hp.required("module/class that will have its self-attention "
                                             "function replaced. For example, in HuggingFace's "
                                             "GPT, the self-attention module is "
                                             "'transformers.models.gpt2.modeling_gpt2.GPT2Attention'")
    attr_to_replace: str = hp.required("attribute that self-attention function will "
                                       "replace. For example, in HuggingFace's "
                                       "GPT2, the self-attention function is '_attn'")
    alibi_attention: str = hp.required("new self-attention function in which ALiBi is "
                                       "implemented. Used to replace "
                                       "'{attention_module}.{attr_to_replace}'")
    mask_replacement_function: Union[str, None] = hp.optional(
        "function to replace model's attention mask. This is "
        "sometimes necessary for evaluating on sequence "
        " lengths longer than the model was initialized to accommodate.",
        default=None)
    heads_per_layer: Union[int, Optional[None]] = hp.optional(
        'Number of attention heads per layer. If '
        '"None", will attempt to determine from model.config.n_head.',
        default=None)
    max_sequence_length: int = hp.optional('Maximum allowable sequence length', default=8192)
    train_sequence_length_scaling: float = hp.optional(
        'Amount by which to scale training sequence length. One batch of training data '
        'will be reshaped from size (sequence_length, batch) to '
        '(sequence_length*train_sequence_length_scaling, batch/train_sequence_length_scaling)',
        default=0.25)

    def initialize_object(self) -> "Alibi":
        return Alibi(**asdict(self))


def apply_alibi(model: torch.nn.Module, heads_per_layer: int, max_sequence_length: int,
                position_embedding_attribute: str, attention_module: torch.nn.Module, attr_to_replace: str,
                alibi_attention: Callable, mask_replacement_function: Union[Callable, None]) -> None:
    """
    Applies ALiBi to the provided model. Removes position embeddings and replaces
        the attention function and attention mask.

    Args:
        model (torch.nn.Module): model to transform
        heads_per_layer (int): number of attention heads per layer
        max_sequence_length (int): maximum sequence length that the
            model will be able to accept without returning an error
        position_embedding_attribute (str): attribute for position
            embeddings. For example in HuggingFace's GPT2, the
            position embeddings are "transformer.wpe".
        attention_module (str): module/class that will have its
            self-attention function replaced. For example, in
            HuggingFace's GPT, the self-attention module is
            transformers.models.gpt2.modeling_gpt2.GPT2Attention.
        attr_to_replace (str): attribute that self-attention function will
            replace. For example, in HuggingFace's GPT2, the
            self-attention function is "_attn".
        alibi_attention (callable): new self-attention function in which
            ALiBi is implemented. Used to replace
            "{attention_module}.{attr_to_replace}".
        mask_replacement_function (callable): function to replace model's
            attention mask. This is sometimes necessary for evaluating
            on sequence lengths longer than the model was initialized to
            accommodate.
    """

    zero_and_freeze_expand_position_embeddings(model=model,
                                               attribute=position_embedding_attribute,
                                               new_embedding_length=max_sequence_length)
    log.info(f" Position embedding expanded to sequence " f"length {max_sequence_length}, zeroed, and frozen")

    def convert_attention(module: torch.nn.Module, module_index: int = None):
        module = register_alibi(module=module, n_heads=heads_per_layer, max_token_length=max_sequence_length)
        setattr(module, attr_to_replace, MethodType(alibi_attention, module))
        if mask_replacement_function:
            module = mask_replacement_function(module, max_sequence_length)
        return module

    transforms = {attention_module: convert_attention}
    replaced_pairs = surgery.replace_module_classes(model, transforms)  # type: ignore

    count = len(replaced_pairs)
    log.info(f" {count} instances of ALiBi added")


class Alibi(Algorithm):
    """
    Algorithm to apply ALiBi to the model. Runs on Event.INIT. This algorithm should be
    applied before the model has been moved to accelerators.

    Args:
        heads_per_layer (int): number of attention heads per layer
        max_sequence_length (int): maximum sequence length that the
            model will be able to accept without returning an error
        position_embedding_attribute (str): attribute for position
            embeddings. For example in HuggingFace's GPT2, the
            position embeddings are "transformer.wpe".
        attention_module_name (str): module/class that will have
            its self-attention function replaced. For example,
            in HuggingFace's GPT, the self-attention module is
            "transformers.models.gpt2.modeling_gpt2.GPT2Attention".
        attr_to_replace (str): attribute that self-attention function will
            replace. For example, in HuggingFace's GPT2, the
            self-attention function is "_attn".
        alibi_attention (str): Path to new self-attention function in which ALiBi is
            implemented. Used to replace "{attention_module}.{attr_to_replace}".
        mask_replacement_function (str): Path to function to replace model's
            attention mask. This is sometimes necessary for evaluating on
            sequence lengths longer than the model was initialized to
            accommodate.
        train_sequence_length_scaling (float): Amount by which to scale
            training sequence length. One batch of training data will be
            reshaped from size (sequence_length, batch) to
            (sequence_length*sequence_length_fraction, batch/sequence_length_fraction).
    """

    def __init__(self, position_embedding_attribute: str, attention_module_name: str, attr_to_replace: str,
                 alibi_attention: str, mask_replacement_function: str, heads_per_layer: int, max_sequence_length: int,
                 train_sequence_length_scaling: float) -> None:

        self.hparams = AlibiHparams(position_embedding_attribute=position_embedding_attribute,
                                    attention_module_name=attention_module_name,
                                    attr_to_replace=attr_to_replace,
                                    alibi_attention=alibi_attention,
                                    mask_replacement_function=mask_replacement_function,
                                    heads_per_layer=heads_per_layer,
                                    max_sequence_length=max_sequence_length,
                                    train_sequence_length_scaling=train_sequence_length_scaling)

    def match(self, event: Event, state: State) -> bool:
        """ Runs on Event.INIT
        """
        return event in (Event.INIT, Event.AFTER_DATALOADER)

    def apply(self, event: Event, state: State, logger: Logger) -> Optional[int]:
        """ Applies ALiBi
        """

        if event == Event.INIT:
            assert state.model is not None

            if "heads_per_layer" not in asdict(self.hparams).keys() or \
            not self.hparams.heads_per_layer:
                try:
                    self.hparams.heads_per_layer = state.model.config.n_head  # type: ignore
                except AttributeError:
                    log.exception("alibi.heads_per_layer not provided, and unable to "
                                  "determine number of heads from model.config.n_head."
                                  " Please provide alibi. heads_per_layer.")

            apply_alibi(
                state.model,
                heads_per_layer=self.hparams.heads_per_layer,  # type: ignore
                max_sequence_length=self.hparams.max_sequence_length,
                position_embedding_attribute=self.hparams.position_embedding_attribute,
                attr_to_replace=self.hparams.attr_to_replace,
                # Access method from string
                attention_module=lazy_import(self.hparams.attention_module_name),
                # Access method from string
                alibi_attention=lazy_import(self.hparams.alibi_attention),
                # Access method from string
                mask_replacement_function=lazy_import(self.hparams.mask_replacement_function))

        elif event == Event.AFTER_DATALOADER:
            # Change sequence length by reshaping data
            if not self.hparams.train_sequence_length_scaling == 1 and \
            hasattr(state, "batch") and isinstance(state.batch, dict):
                sequence_scaling = self.hparams.train_sequence_length_scaling
                for k, v in state.batch.items():
                    batch_len, sequence_len = v.shape[0], v.shape[1]
                    state.batch[k] = v.reshape(int(batch_len / sequence_scaling), int(sequence_len * sequence_scaling))


def zero_and_freeze_expand_position_embeddings(model: torch.nn.Module, new_embedding_length: int, attribute: str):
    try:
        pos_embedding_module = attrgetter(attribute)(model)
        old_weight = getattr(pos_embedding_module, "weight")
        new_weight = torch.nn.Parameter(
            torch.zeros((new_embedding_length, old_weight.shape[1]),
                        dtype=old_weight.dtype,
                        layout=old_weight.layout,
                        device=old_weight.device))
        new_weight.requires_grad = False
        setattr(pos_embedding_module, "weight", new_weight)
    except AttributeError:
        log.error(f"Unable to zero and freeze position embeddings. Model "
                  f"{model} may lack attribute {attribute}, or position "
                  f"embeddings may lack attribute 'weight'.")


def register_alibi(module: torch.nn.Module, n_heads: int, max_token_length: int):
    # Modified from https://github.com/ofirpress/attention_with_linear_biases/blob/master/fairseq/models/transformer.py#L742
    slopes = torch.Tensor(get_alibi_head_slopes(n_heads))
    # In the next line, the part after the * is what constructs the diagonal matrix
    # (right matrix in Figure 3 in the paper).
    # If you run it you'll see that it doesn't exactly print out the same matrix as we
    # have in Figure 3, but one where all rows are identical.
    # This works because the softmax operation is invariant to translation, and our bias
    # functions are always linear.
    alibi = slopes.unsqueeze(1).unsqueeze(1) * \
        torch.arange(max_token_length). \
        unsqueeze(0).unsqueeze(0).expand(n_heads, -1, -1)
    module.register_buffer("alibi", alibi)
    return module


def get_alibi_head_slopes(n_heads: int):

    def get_slopes_power_of_2(n_heads):
        start = (2**(-2**-(math.log2(n_heads) - 3)))
        ratio = start
        return [start * ratio**i for i in range(n_heads)]

    # In the paper, they only train models that have 2^a heads for some a. This function
    # has some good properties that only occur when the input is a power of 2. To
    # maintain that even when the number of heads is not a power of 2, we use a
    # workaround.
    if math.log2(n_heads).is_integer():
        return get_slopes_power_of_2(n_heads)
    else:
        closest_power_of_2 = 2**math.floor(math.log2(n_heads))
        return get_slopes_power_of_2(closest_power_of_2) + get_alibi_head_slopes(
            2 * closest_power_of_2)[0::2][:n_heads - closest_power_of_2]


def lazy_import(name: Union[str, None]) -> Any[Callable, ModuleType, None]:
    if not name:
        return None
    components = name.split('.')
    try:
        mod = importlib.import_module(components[0])
    except (ValueError, ModuleNotFoundError):
        log.exception(f"Module {components[0]} not found when attempting "
                      f"to import {name}. Please confirm the name and "
                      f"module path you're attempting to import.")
    try:
        mod = attrgetter('.'.join(components[1:]))(mod)  # type: ignore
    except (ValueError, AttributeError):
        log.exception(f"Unable to import {name}. "
                      f"Please confirm the name and module "
                      f" path you're attempting to import.")
    return mod  # type: ignore



# Copyright 2021 MosaicML. All Rights Reserved.

from __future__ import annotations

import multiprocessing
from abc import abstractmethod
from dataclasses import dataclass

import torch
import yahp as hp

from composer.trainer.devices.device import Device
from composer.trainer.devices.device_cpu import DeviceCPU
from composer.trainer.devices.device_gpu import DeviceGPU


@dataclass
class DeviceHparams(hp.Hparams):

    @abstractmethod
    def initialize_object(self) -> Device:
        pass


@dataclass
class GPUDeviceHparams(DeviceHparams):
    n_gpus: int = hp.required(doc="Number of GPUs to use per node. Set to -1 to use all.", template_default=-1)
    prefetch_in_cuda_stream: bool = hp.optional(doc="Whether to use a separate cuda stream for prefetching",
                                                default=True)

    def initialize_object(self) -> DeviceGPU:
        if self.n_gpus == -1:
            self.n_gpus = torch.cuda.device_count()
        if self.n_gpus < 1:
            raise ValueError("When using a GPU device, n_gpus must be at least 1.")
        return DeviceGPU(n_gpus=self.n_gpus, prefetch_in_cuda_stream=self.prefetch_in_cuda_stream)


@dataclass
class CPUDeviceHparams(DeviceHparams):
    n_cpus: int = hp.required(doc="Number of CPUs to use per node. Set to -1 to use multiprocessing.cpu_count()",
                              template_default=-1)

    def initialize_object(self) -> DeviceCPU:
        if self.n_cpus == -1:
            self.n_cpus = multiprocessing.cpu_count()
        if self.n_cpus < 1:
            raise ValueError("When using a CPU device, n_cpus must be at least 1.")
        return DeviceCPU(num_cpus=self.n_cpus)



# Copyright 2021 MosaicML. All Rights Reserved.

from __future__ import annotations

from typing import TYPE_CHECKING, Protocol

if TYPE_CHECKING:
    from composer.core.types import StateDict


class Serializable(Protocol):
    """Interface for serialization; used by checkpointing."""

    def state_dict(self) -> StateDict:
        """state_dict() returns a dictionary representing the internal state.
        The returned dictionary must be pickale-able via :meth:`torch.save()`.
    
        Returns:
            StateDict: The state of the object
        """
        return {}

    def load_state_dict(self, state: StateDict) -> None:
        """load_state_dict() restores the state of the object

        Args:
            state (StateDict): The state of the object,
                as previously returned by :meth:`.state_dict`
        """
        pass



# Copyright 2021 MosaicML. All Rights Reserved.

from __future__ import annotations

import logging
from contextlib import contextmanager
from typing import Generator, Optional, Union

import torch

from composer.core.state import State
from composer.core.types import Batch, DataLoader, Precision, StateDict, Tensor, TPrefetchFn
from composer.datasets.dataloader import WrappedDataLoader
from composer.trainer.devices.device import Device, T_nnModule

logger = logging.getLogger(__name__)


class PrefetchedDataLoader(WrappedDataLoader):

    def __init__(self, dataloader: DataLoader, batch_preprocessing_fn: Optional[TPrefetchFn]) -> None:
        super().__init__(dataloader)
        self.device = torch.device("cpu")
        self.batch_preprocessing_fn = batch_preprocessing_fn

    def __iter__(self) -> Generator[Batch, None, None]:
        for batch in self.dataloader:
            if self.batch_preprocessing_fn is None:
                yield batch
            else:
                yield self.batch_preprocessing_fn(batch)


class DeviceCPU(Device):

    def __init__(self, num_cpus: int):
        self._num_cpus = num_cpus

    @property
    def nproc_per_node(self) -> int:
        return self._num_cpus

    def prepare(self, state: State) -> None:
        logger.info("Preparing CPU worker")
        # No preperation required
        return

    def module_to_device(self, module: T_nnModule) -> T_nnModule:
        return module

    def dataloader_to_device(self, dataloader: DataLoader, prefetch_fn: Optional[TPrefetchFn]) -> DataLoader:
        return PrefetchedDataLoader(dataloader, prefetch_fn)

    def tensor_to_device(self, tensor: Tensor) -> Tensor:
        return tensor

    @contextmanager
    def precision_context(self, precision: Union[str, Precision]) -> Generator[None, None, None]:
        precision = Precision(precision)
        if precision == Precision.FP32:
            yield
        else:
            raise ValueError(f"Precision {precision} not supported for a CPU")

    @property
    def ddp_backend(self) -> str:
        return "gloo"

    def state_dict(self) -> StateDict:
        # CPU device has no RNG state
        return {}

    def load_state_dict(self, state: StateDict) -> None:
        if len(state) != 0:
            raise ValueError("CPU device has no state.")



# Copyright 2021 MosaicML. All Rights Reserved.

import os

import pytest
import torch.distributed as dist
from _pytest.monkeypatch import MonkeyPatch

from composer.core.logging import Logger, LogLevel
from composer.core.state import State
from composer.loggers.file_logger import FileLoggerBackend
from composer.loggers.logger_hparams import FileLoggerBackendHparams


@pytest.fixture
def log_file_name(ddp_tmpdir: str) -> str:
    return os.path.join(ddp_tmpdir, "output.log")


@pytest.fixture
def log_destination(log_file_name: str) -> FileLoggerBackend:
    return FileLoggerBackendHparams(
        every_n_batches=3,
        every_n_epochs=2,
        log_level=LogLevel.BATCH,
        filename=log_file_name,
        buffer_size=1,
        flush_every_n_batches=1,
    ).initialize_object()


def test_file_logger(dummy_state: State, log_destination: FileLoggerBackend, monkeypatch: MonkeyPatch,
                     log_file_name: str):
    dummy_state.step = 2
    dummy_state.epoch = 2
    logger = Logger(dummy_state, backends=[log_destination])
    monkeypatch.setattr(dist, "get_rank", lambda: 0)
    log_destination.training_start(dummy_state, logger)
    logger.metric_fit({"metric": "fit"})  # should print
    logger.metric_epoch({"metric": "epoch"})  # should print
    logger.metric_batch({"metric": "batch"})  # should print
    logger.metric_verbose({"metric": "verbose"})  # should NOT print, since we're on the BATCH log level
    dummy_state.epoch = 3
    logger.metric_epoch({"metric": "epoch1"})  # should NOT print, since we print every 2 epochs
    dummy_state.epoch = 4
    dummy_state.step = 3
    log_destination.batch_end(dummy_state, logger)
    logger.metric_epoch({"metric": "epoch2"})  # should print
    logger.metric_batch({"metric": "batch1"})  # should NOT print, since we print every 3 steps
    log_destination.batch_end(dummy_state, logger)
    log_destination.training_end(dummy_state, logger)
    with open(log_file_name, 'r') as f:
        assert f.readlines() == [
            '[FIT][step=2]: { "metric": "fit", }\n',
            '[EPOCH][step=2]: { "metric": "epoch", }\n',
            '[BATCH][step=2]: { "metric": "batch", }\n',
            '[EPOCH][step=3]: { "metric": "epoch2", }\n',
        ]


class TestCoreLogger:

    @pytest.mark.parametrize("rank", [0, 1])
    def test_deferred(self, dummy_state_without_rank: State, log_file_name: str, monkeypatch: MonkeyPatch,
                      log_destination: FileLoggerBackend, rank: int):
        dummy_state = dummy_state_without_rank
        dummy_state.step = 2
        dummy_state.epoch = 0
        logger = Logger(dummy_state, backends=[log_destination])
        logger.metric_batch({"metric": "before_training_start"})
        monkeypatch.setattr(dist, "get_rank", lambda: rank)
        log_destination.training_start(dummy_state, logger)
        logger.metric_batch({"metric": "after_training_start"})
        log_destination.batch_end(dummy_state, logger)
        log_destination.training_end(dummy_state, logger)
        if rank == 0:
            with open(log_file_name, 'r') as f:
                assert f.readlines() == [
                    '[BATCH][step=2]: { "metric": "before_training_start", }\n',
                    '[BATCH][step=2]: { "metric": "after_training_start", }\n',
                ]
            return
        else:
            assert rank == 1
            assert not os.path.exists(log_file_name), "nothing should be logged on rank 1"

    def test_deep_copy(self, dummy_state_without_rank: State, log_destination: FileLoggerBackend,
                       monkeypatch: MonkeyPatch, log_file_name: str):
        # This test ensures that the logger deepcopies the logged metric when using deferred logging
        dummy_state = dummy_state_without_rank
        dummy_state.step = 2
        dummy_state.epoch = 0
        logger = Logger(dummy_state, backends=[log_destination])
        metric_data = [["hello"]]
        logger.metric_batch({"metric": metric_data})
        metric_data[0] = ["world"]
        monkeypatch.setattr(dist, "get_rank", lambda: 0)
        log_destination.training_start(dummy_state, logger)
        logger.metric_batch({"metric": metric_data})
        log_destination.batch_end(dummy_state, logger)
        log_destination.training_end(dummy_state, logger)
        with open(log_file_name, 'r') as f:
            assert f.readlines() == [
                '[BATCH][step=2]: { "metric": [["hello"]], }\n',
                '[BATCH][step=2]: { "metric": [["world"]], }\n',
            ]



# Copyright 2021 MosaicML. All Rights Reserved.

from __future__ import annotations

import logging
from dataclasses import asdict, dataclass
from typing import Optional

import yahp as hp

from composer.algorithms.algorithm_hparams import AlgorithmHparams

log = logging.getLogger(__name__)


@dataclass
class SWAHparams(AlgorithmHparams):

    swa_start: float = hp.optional(
        doc='Percentage of epochs before starting to apply SWA. Default 0.8.',
        default=0.8,
    )
    anneal_epochs: int = hp.optional(
        doc='Number of annealing epochs. Default 10.',
        default=10,
    )
    swa_lr: Optional[float] = hp.optional(
        doc='The final learning rate to anneal towards with this scheduler. '
        'Set to None for no annealing. Default: None',
        default=None,
    )

    def initialize_object(self):
        from composer.algorithms.swa import SWA
        return SWA(**asdict(self))



# Copyright 2021 MosaicML. All Rights Reserved.

from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import asdict, dataclass
from typing import TYPE_CHECKING, Any, Dict, List, Optional

import yahp as hp

from composer.core.logging import BaseLoggerBackend, LogLevel
from composer.core.types import JSON

if TYPE_CHECKING:
    from composer.loggers.file_logger import FileLoggerBackend
    from composer.loggers.tqdm_logger import TQDMLoggerBackend
    from composer.loggers.wandb_logger import WandBLoggerBackend


@dataclass
class BaseLoggerBackendHparams(hp.Hparams, ABC):

    @abstractmethod
    def initialize_object(self, config: Optional[Dict[str, Any]] = None) -> BaseLoggerBackend:
        """ Initializes a Base Logger

        Args:
            config: An optional dictionary that a logger can use on initialization to
                log a config or configure the logger
        """
        pass


@dataclass
class FileLoggerBackendHparams(BaseLoggerBackendHparams):
    log_level: LogLevel = hp.optional("The maximum verbosity to log. Default: EPOCH", default=LogLevel.EPOCH)
    filename: str = hp.optional("The path to the logfile. Can also be `stdout` or `stderr`. Default: stdout",
                                default="stdout")
    buffer_size: int = hp.optional("Number of bytes to buffer. Defaults to 1 for line-buffering. "
                                   "See https://docs.python.org/3/library/functions.html#open",
                                   default=1)  # line buffering. Python's default is -1.
    flush_every_n_batches: int = hp.optional(
        "Even if the buffer is not full, write to the file after this many steps. "
        "Defaults to 1 (every step).",
        default=1)
    every_n_epochs: int = hp.optional(
        "Frequency of logging messages for messages of LogLevel.EPOCH and higher."
        "Defaults to 1 (every epoch).",
        default=1)
    every_n_batches: int = hp.optional(
        "Frequency of logging messages for messages of LogLevel.BATCH and higher."
        "Defaults to 1 (every batch).",
        default=1)

    def initialize_object(self, config: Optional[Dict[str, Any]] = None) -> FileLoggerBackend:

        from composer.loggers.file_logger import FileLoggerBackend
        return FileLoggerBackend(**asdict(self), config=config)


@dataclass
class WandBLoggerBackendHparams(BaseLoggerBackendHparams):
    """
    WandBOutputConfig configures logging to Weights and Biases (wandb.ai)

    Args:
        init_params: kwargs to pass into `wandb.init()`
    """

    project: Optional[str] = hp.optional(doc="wandb project name", default=None)
    name: Optional[str] = hp.optional(doc="wandb run name", default=None)
    entity: Optional[str] = hp.optional(doc="wandb entity", default=None)
    tags: str = hp.optional(doc="wandb tags comma separated", default="")
    extra_init_params: Dict[str, JSON] = hp.optional(doc="wandb parameters", default_factory=dict)

    def initialize_object(self, config: Optional[Dict[str, Any]] = None) -> WandBLoggerBackend:
        tags = list(set([x.strip() for x in self.tags.split(",") if x.strip() != ""]))

        if config is not None:
            if "algorithms" in config:
                algos_dict_list = config.get("algorithms", [])
                algorithm_names = []
                for algo_dict in algos_dict_list:
                    if len(algo_dict.keys()) != 0:
                        algorithm_names.append(list(algo_dict.keys())[0])

                tags += algorithm_names
                algorithms_str = "-".join(algorithm_names)
                config["algorithms_str"] = algorithms_str

                # TODO(@averylamp): Remove special algos config columns after data_collection
                # This format is to keep runs standardized for data collection runs
                for algo in algos_dict_list:  # type: ignore
                    for algo_name, algo_params in algo.items():
                        for param_name, param_value in algo_params.items():
                            name = ".".join(["algo", algo_name, param_name])
                            config[name] = param_value

            if "model" in config:
                model_dict = config.get("model", {"unknown": None})
                model_name = "unknown"
                if len(model_dict) == 1:
                    model_name = list(model_dict.keys())[0]
                config["model"] = model_name

            def get_flattened_dict(data: Dict[str, Any], _prefix: List[str] = []) -> Dict[str, Any]:
                """
                Flattens a dictionary with list or sub dicts to have dot syntax

                i.e. {
                  "sub_dict":{
                    "sub_list":[
                      "sub_sub_dict":{
                        "field1": 12,
                        "field2": "tomatoes"
                      }
                    ]
                  },
                  "field3": "potatoes"
                }

                returns:
                {
                  "sub_dict.sub_list.sub_sub_dict.field1": 12,
                  "sub_dict.sub_list.sub_sub_dict.field2": "tomatoes,
                  "field3": "potatoes",
                }
                """
                all_items = dict()
                for key, val in data.items():
                    key_items = _prefix + [key]
                    key_name = ".".join(key_items)
                    if isinstance(val, dict):
                        all_items.update(get_flattened_dict(val, key_items))
                    elif isinstance(val, list):
                        found_sub_dicts = False
                        for item in val:
                            if isinstance(item, dict):
                                found_sub_dicts = True
                                for sub_key, sub_val in item.items():
                                    all_items.update(get_flattened_dict(sub_val, key_items + [sub_key]))
                        if not found_sub_dicts:
                            all_items[key_name] = val
                    else:
                        all_items[key_name] = val
                return all_items

            flattened_config = get_flattened_dict(data=config)
            if "config" not in self.extra_init_params:
                self.extra_init_params["config"] = {}
            self.extra_init_params["config"].update(flattened_config)  # type: ignore

        kwargs = {
            "project": self.project,
            "name": self.name,
            "entity": self.entity,
            "tags": tags,
        }

        kwargs.update(self.extra_init_params)

        from composer.loggers.wandb_logger import WandBLoggerBackend
        return WandBLoggerBackend(**kwargs)


@dataclass
class TQDMLoggerBackendHparams(BaseLoggerBackendHparams):

    def initialize_object(self, config: Optional[Dict[str, Any]] = None) -> TQDMLoggerBackend:
        from composer.loggers.tqdm_logger import TQDMLoggerBackend
        return TQDMLoggerBackend(config=config)



# Copyright 2021 MosaicML. All Rights Reserved.

import contextlib
import logging
from typing import Any, Optional, Tuple

import torch
import torch.nn as nn
from monai.losses import DiceLoss

from composer.core.types import BatchPair, Metrics, Tensor, Tensors
from composer.models.base import BaseMosaicModel
from composer.models.loss import Dice
from composer.models.unet.model import UNet as UNetModel
from composer.models.unet.unet_hparams import UnetHparams

log = logging.getLogger(__name__)


class UNet(BaseMosaicModel):

    n_classes: Optional[int] = None

    def __init__(self, hparams: UnetHparams) -> None:
        super().__init__()
        self.hparams = hparams
        self.module = self.build_nnunet()

        self.dice = Dice(3)

        self.dloss = DiceLoss(include_background=False, softmax=True, to_onehot_y=True, batch=True)
        self.closs = nn.CrossEntropyLoss()

    def loss(self, outputs: Any, batch: BatchPair) -> Tensors:

        _, y = batch
        y = y.squeeze(1)  # type: ignore

        assert isinstance(y, Tensor)

        loss = self.dloss(outputs, y)
        loss += self.closs(outputs, y[:, 0].long())
        return loss

    @staticmethod
    def metric_mean(name, outputs):
        return torch.stack([out[name] for out in outputs]).mean(dim=0)

    def metrics(self, train: bool = False) -> Metrics:

        return self.dice

    def forward(self, batch: BatchPair) -> Tensor:
        x, y = batch
        context = contextlib.nullcontext if self.training else torch.no_grad

        x = x.squeeze(1)  # type: ignore

        with context():
            logits = self.module(x)

        return logits

    def inference2d(self, image):
        """Runs inference on a 3D image, by passing each depth slice through the model.
        """
        batch_modulo = image.shape[2] % 64
        if batch_modulo != 0:
            batch_pad = 64 - batch_modulo
            image = nn.ConstantPad3d((0, 0, 0, 0, batch_pad, 0), 0)(image)

        image = torch.transpose(image.squeeze(0), 0, 1)
        preds_shape = (image.shape[0], 4, *image.shape[2:])
        preds = torch.zeros(preds_shape, dtype=image.dtype, device=image.device)
        for start in range(0, image.shape[0] - 64 + 1, 64):
            end = start + 64
            with torch.no_grad():
                pred = self.module(image[start:end])
            preds[start:end] = pred.data
        if batch_modulo != 0:
            preds = preds[batch_pad:]  # type: ignore
        return torch.transpose(preds, 0, 1).unsqueeze(0)

    def validate(self, batch: BatchPair) -> Tuple[Any, Any]:
        assert self.training is False, "For validation, model must be in eval mode"

        img, lbl = batch
        pred = self.inference2d(img)
        return pred, lbl[:, 0].long()  # type: ignore

    def build_nnunet(self) -> torch.nn.Module:
        kernels = [[3, 3]] * 6
        strides = [[1, 1]] + [[2, 2]] * 5

        model = UNetModel(in_channels=4,
                          n_class=4,
                          kernels=kernels,
                          strides=strides,
                          dimension=2,
                          residual=True,
                          normalization_layer="instance",
                          negative_slope=0.01)

        return model



# Copyright 2021 MosaicML. All Rights Reserved.

from __future__ import annotations

from typing import TYPE_CHECKING, Optional, Tuple

import torch
import torch.nn.functional as F
from torch import Tensor
from torchmetrics import Metric
from torchmetrics.utilities.data import to_categorical

if TYPE_CHECKING:
    from composer.core.types import Tensor


class Dice(Metric):

    def __init__(self, nclass):
        super().__init__(dist_sync_on_step=True)
        self.add_state("n_updates", default=torch.zeros(1), dist_reduce_fx="sum")
        self.add_state("dice", default=torch.zeros((nclass,)), dist_reduce_fx="sum")

    def update(self, pred, target):
        self.n_updates += 1  # type: ignore
        self.dice += self.compute_stats(pred, target)

    def compute(self):
        dice = 100 * self.dice / self.n_updates  # type: ignore
        best_sum_dice = dice[:]
        top_dice = round(torch.mean(best_sum_dice).item(), 2)
        return top_dice

    @staticmethod
    def compute_stats(pred, target):
        num_classes = pred.shape[1]
        scores = torch.zeros(num_classes - 1, device=pred.device, dtype=torch.float32)
        for i in range(1, num_classes):
            if (target != i).all():
                # no foreground class
                _, _pred = torch.max(pred, 1)
                scores[i - 1] += 1 if (_pred != i).all() else 0
                continue
            _tp, _fp, _tn, _fn, _ = _stat_scores(pred, target, class_index=i)  # type: ignore
            denom = (2 * _tp + _fp + _fn).to(torch.float)
            score_cls = (2 * _tp).to(torch.float) / denom if torch.is_nonzero(denom) else 0.0
            scores[i - 1] += score_cls
        return scores


def _stat_scores(
    preds: Tensor,
    target: Tensor,
    class_index: int,
    argmax_dim: int = 1,
) -> Tuple[Tensor, Tensor, Tensor, Tensor, Tensor]:

    if preds.ndim == target.ndim + 1:
        preds = to_categorical(preds, argmax_dim=argmax_dim)

    tp = ((preds == class_index) * (target == class_index)).to(torch.long).sum()
    fp = ((preds == class_index) * (target != class_index)).to(torch.long).sum()
    tn = ((preds != class_index) * (target != class_index)).to(torch.long).sum()
    fn = ((preds != class_index) * (target == class_index)).to(torch.long).sum()
    sup = (target == class_index).to(torch.long).sum()

    return tp, fp, tn, fn, sup


def _infer_target_type(input: Tensor, targets: Tensor) -> str:
    """
    Attempts to infer whether target is indicies format (e.g. [1, 4, 7]) or
    one_hot format (e.g. [[0, 1, 0], [1, 0, 0], ...])
    """
    if input.shape == targets.shape:
        return 'one_hot'
    elif input.ndim == targets.ndim + 1:
        return 'indicies'
    else:
        raise RuntimeError(f'Unable to infer indicies or one_hot. Targets has shape {targets.shape}'
                           f' and the inputs to cross entropy has shape {input.shape}. For one_hot, '
                           'expect targets.shape == inputs.shape. For indicies, expect '
                           'inputs.ndim == targets.ndim + 1')


def ensure_targets_one_hot(input: Tensor, targets: Tensor) -> Tensor:
    if _infer_target_type(input, targets) == 'indicies':
        targets = F.one_hot(targets, num_classes=input.shape[1])
    return targets


def soft_cross_entropy(input: Tensor,
                       target: Tensor,
                       weight: Optional[Tensor] = None,
                       size_average: Optional[bool] = None,
                       ignore_index: int = -100,
                       reduce: Optional[bool] = None,
                       reduction: str = 'mean'):
    """
    Drop-in replacement for torch.CrossEntropy that can handle dense labels.
    This function will be obsolete with https://github.com/pytorch/pytorch/pull/61044
    """
    target_type = _infer_target_type(input, target)

    if target_type == 'indicies':
        return F.cross_entropy(input, target, weight, size_average, ignore_index, reduce, reduction)
    elif target_type == 'one_hot':
        assert reduction in ['sum', 'mean', 'none'], f"{reduction} reduction not supported."
        assert size_average is None, "size_average is deprecated"
        assert reduce is None, "reduce is deprecated"
        assert ignore_index == -100, "ignore_index not supported."
        probs = -1 * (target * F.log_softmax(input, dim=1))

        if weight is not None:
            probs *= weight / weight.sum()  # allow broadcast along batch dim

        probs = probs.sum(dim=1)

        if reduction == 'sum':
            probs = probs.sum(dim=0)
        elif reduction == 'mean':
            probs = probs.mean(dim=0)

        return probs
    else:
        raise ValueError(f"Unrecognized target type {target_type}")


class CrossEntropyLoss(Metric):
    """
    Torchmetric implementation to calculate validation loss
    """

    def __init__(self, dist_sync_on_step=False):
        super().__init__(dist_sync_on_step=dist_sync_on_step)

        self.add_state("sum_loss", default=torch.tensor(0.), dist_reduce_fx="sum")
        self.add_state("total_batches", default=torch.tensor(0), dist_reduce_fx="sum")

    def update(self, preds: Tensor, target: Tensor) -> None:
        # Loss calculated over samples/batch, accumulate loss over all batches
        self.sum_loss += soft_cross_entropy(preds, target)
        assert isinstance(self.total_batches, Tensor)
        self.total_batches += 1

    def compute(self) -> Tensor:
        # Return average loss over entire validation dataset
        assert isinstance(self.total_batches, Tensor)
        assert isinstance(self.sum_loss, Tensor)
        return self.sum_loss / self.total_batches



# Copyright 2021 MosaicML. All Rights Reserved.

from __future__ import annotations

import logging
from dataclasses import asdict, dataclass
from typing import Union

import torch
import yahp as hp
from PIL.Image import Image
from torchvision.transforms import functional as TF

from composer.algorithms import AlgorithmHparams
from composer.core import Algorithm, Event, Logger, State
from composer.core.types import Tensor
from composer.utils.data import add_dataset_transform

log = logging.getLogger(__name__)


def colout(img: Union[torch.Tensor, Image], p_row: float, p_col: float) -> Union[torch.Tensor, Image]:
    """Drops random rows and columns from a single image.

    Args:
        img (torch.Tensor or PIL Image): An input image as a torch.Tensor or PIL image
        p_row (float): Fraction of rows to drop (drop along H).
        p_col (float): Fraction of columns to drop (drop along W).

    Returns:
        torch.Tensor or PIL Image: A smaller image with rows and columns dropped
    """

    as_PIL = False

    # Convert image to Tensor if needed
    if isinstance(img, Image):
        as_PIL = True
        img = TF.to_tensor(img)

    if not isinstance(img, torch.Tensor):
        raise ValueError("Invalid input type: img must be either torch.Tensor or PIL Image.")

    # Get the dimensions of the image
    row_size = img.shape[1]
    col_size = img.shape[2]

    # Determine how many rows and columns to keep
    kept_row_size = int((1 - p_row) * row_size)
    kept_col_size = int((1 - p_col) * col_size)

    # Randomly choose indices to keep. Must be sorted for slicing
    kept_row_idx = sorted(torch.randperm(row_size)[:kept_row_size].numpy())
    kept_col_idx = sorted(torch.randperm(col_size)[:kept_col_size].numpy())

    # Keep only the selected row and columns
    img = img[:, kept_row_idx, :]
    img = img[:, :, kept_col_idx]

    # Convert back to PIL for the rest of the augmentation pipeline
    if as_PIL:
        return TF.to_pil_image(img)
    else:
        return img


class ColOutTransform:

    def __init__(self, p_row: float, p_col: float):
        """ Torchvision-like transform for performing the ColOut augmentation, where random rows and columns are
            dropped from a single image.

        Args:
            p_row (float): Fraction of rows to drop (drop along H).
            p_col (float): Fraction of columns to drop (drop along W).
        """
        self.p_row = p_row
        self.p_col = p_col

    def __call__(self, img: Union[torch.Tensor, Image]) -> Union[torch.Tensor, Image]:
        """ Drops random rows and columns from a single image.

        Args:
            img (torch.Tensor or PIL Image): An input image as a torch.Tensor or PIL image

        Returns:
            torch.Tensor or PIL Image: A smaller image with rows and columns dropped
        """
        return colout(img, self.p_row, self.p_col)


def batch_colout(X: torch.Tensor, p_row: float, p_col: float) -> torch.Tensor:
    """ Implements ColOut augmentation on the batch level, where random rows and columns are
        dropped from all images in a batch.

    Args:
        X (Tensor): Batch Tensor image of size (N, C, H, W).
        p_row: Fraction of rows to drop (drop along H).
        p_col: Fraction of columns to drop (drop along W).

    Returns:
        torch.Tensor: Input batch tensor with randomly dropped columns and rows.
    """

    # Get the dimensions of the image
    row_size = X.shape[2]
    col_size = X.shape[3]

    # Determine how many rows and columns to keep
    kept_row_size = int((1 - p_row) * row_size)
    kept_col_size = int((1 - p_col) * col_size)

    # Randomly choose indices to keep. Must be sorted for slicing
    kept_row_idx = sorted(torch.randperm(row_size)[:kept_row_size].numpy())
    kept_col_idx = sorted(torch.randperm(col_size)[:kept_col_size].numpy())

    # Keep only the selected row and columns
    X_colout = X[:, :, kept_row_idx, :]
    X_colout = X_colout[:, :, :, kept_col_idx]
    return X_colout


@dataclass
class ColOutHparams(AlgorithmHparams):
    p_row: float = hp.optional(doc="Fraction of rows to drop. Default: 0.15", default=0.15)
    p_col: float = hp.optional(doc="Fraction of cols to drop. Default: 0.15", default=0.15)
    batch: bool = hp.optional(doc="Run ColOut at the batch level", default=True)

    def initialize_object(self) -> ColOut:
        return ColOut(**asdict(self))


class ColOut(Algorithm):

    def __init__(self, p_row: float = 0.15, p_col: float = 0.15, batch: bool = True):
        """ Algorithm interface for the ColOut augmentation, where random rows and columns are
            dropped from input images.

        Args:
            p_row (float, optional): Fraction of rows to drop (drop along H). Defaults to 0.15.
            p_col (float, optional): Fraction of columns to drop (drop along W). Defaults to 0.15.
            batch (bool, optional): Run ColOut at the batch level. Defaults to True.
        """
        if not (0 <= p_col <= 1):
            raise ValueError("p_col must be between 0 and 1")

        if not (0 <= p_row <= 1):
            raise ValueError("p_row must be between 0 and 1")

        self.hparams = ColOutHparams(p_row, p_col, batch)

    def match(self, event: Event, state: State) -> bool:
        """ Apply on Event.TRAINING_START for samplewise or Event.AFTER_DATALOADER for batchwise """
        if self.hparams.batch:
            return event == Event.AFTER_DATALOADER
        else:
            return event == Event.TRAINING_START

    def _apply_sample(self, state: State) -> None:
        """ Add the ColOut dataset transform to the dataloader """
        assert state.train_dataloader is not None
        dataset = state.train_dataloader.dataset

        transform = ColOutTransform(p_row=self.hparams.p_row, p_col=self.hparams.p_col)

        if hasattr(dataset, "transform"):
            add_dataset_transform(dataset, transform)
        else:
            raise ValueError(
                f"Dataset of type {type(dataset)} has no attribute 'transform'. Expected TorchVision dataset.")

    def _apply_batch(self, state: State) -> None:
        """ Transform a batch of images using the ColOut augmentation """
        inputs, labels = state.batch_pair
        assert isinstance(inputs, Tensor), "Multiple Tensors not supported yet for ColOut"
        new_inputs = batch_colout(inputs, p_row=self.hparams.p_row, p_col=self.hparams.p_col)

        state.batch = (new_inputs, labels)

    def apply(self, event: Event, state: State, logger: Logger) -> None:
        """ Apply the relevant ColOut augmentation

        Args:
            event (Event): the current event
            state (State): the current trainer state
            logger (Optional[Logger], optional): the training logger. Defaults to None.
        """
        if self.hparams.batch:
            self._apply_batch(state)
        else:
            self._apply_sample(state)



# Copyright 2021 MosaicML. All Rights Reserved.

import logging
from dataclasses import asdict, dataclass
from typing import Optional

import numpy as np
import torch
import yahp as hp

from composer.algorithms import AlgorithmHparams
from composer.core.types import Algorithm, Event, Logger, State, Tensor

log = logging.getLogger(__name__)


def generate_mask(mask: Tensor, width: int, height: int, x: int, y: int, cutout_length: int) -> Tensor:
    y1 = np.clip(y - cutout_length // 2, 0, height)
    y2 = np.clip(y + cutout_length // 2, 0, height)
    x1 = np.clip(x - cutout_length // 2, 0, width)
    x2 = np.clip(x + cutout_length // 2, 0, width)

    mask[:, :, y1:y2, x1:x2] = 0.

    return mask


def apply_cutout(X: Tensor, mask: Tensor):
    return X * mask


def cutout(X: Tensor, n_holes: int, length: int) -> Tensor:
    """
    Implements CutOut augmentation from https://arxiv.org/abs/1708.04552 on the batch level.
    Adapted from the implementation in https://github.com/uoguelph-mlrg/Cutout

    Args:
        X (Tensor): Batch Tensor image of size (B, C, H, W).
        n_holes: Integer number of holes to cut out
        length: Side length of the square hole to cut out.

    Returns:
        X_cutout: Image with n_holes of dimension length x length cut out of it.
    """
    h = X.size(2)
    w = X.size(3)

    mask = torch.ones_like(X)
    for n in range(n_holes):
        y = np.random.randint(h)
        x = np.random.randint(w)

        mask = generate_mask(mask, w, h, x, y, length)

    X_cutout = apply_cutout(X, mask)
    return X_cutout


@dataclass
class CutOutHparams(AlgorithmHparams):

    n_holes: int = hp.required('Number of holes to cut out', template_default=1)
    length: int = hp.required('Side length of the square hole to cut out', template_default=112)

    def initialize_object(self) -> "CutOut":
        return CutOut(**asdict(self))


class CutOut(Algorithm):
    """
    Implements CutOut augmentation from https://arxiv.org/abs/1708.04552 on the batch level.


    Args:
        X (Tensor): Batch Tensor image of size (B, C, H, W).
        n_holes: Integer number of holes to cut out
        length: Side length of the square hole to cut out.
    """

    def __init__(self, n_holes: int, length: int):
        self.hparams = CutOutHparams(n_holes=n_holes, length=length)

    def match(self, event: Event, state: State) -> bool:
        return event == Event.AFTER_DATALOADER

    def apply(self, event: Event, state: State, logger: Logger) -> Optional[int]:
        x, y = state.batch_pair
        assert isinstance(x, Tensor), "Multiple tensors not supported for Cutout."

        new_x = cutout(X=x, **asdict(self.hparams))
        state.batch = (new_x, y)



# Copyright 2021 MosaicML. All Rights Reserved.

from composer.utils import augmentation_primitives as augmentation_primitives
from composer.utils.ddp import get_global_rank as get_global_rank
from composer.utils.ddp import is_rank_set as is_rank_set
from composer.utils.ddp import is_rank_zero as is_rank_zero
from composer.utils.determinism import get_random_seed as get_random_seed
from composer.utils.determinism import seed_all as seed_all
from composer.utils.iter_helpers import ensure_tuple as ensure_tuple
from composer.utils.iter_helpers import map_collection as map_collection
from composer.utils.iter_helpers import zip_collection as zip_collection
from composer.utils.string_enum import StringEnum as StringEnum
from composer.utils.tables import assert_attributes_exist as assert_attributes_exist
from composer.utils.tables import build_markdown_table as build_markdown_table
from composer.utils.tables import get_metadata as get_metadata
from composer.utils.tables import list_dirs as list_dirs
from composer.utils.tables import update_table_in_file as update_table_in_file
from composer.utils.tensor import make_empty_tensor as make_empty_tensor



# Copyright 2021 MosaicML. All Rights Reserved.

import torch

from composer.models.efficientnets import EfficientNet


def test_efficientb0_activate_shape():
    random_input = torch.rand(2, 3, 224, 224)

    model = EfficientNet.get_model_from_name(
        'efficientnet-b0',
        num_classes=1000,
        drop_connect_rate=0.2,
    )
    # Test Stem
    out = model.conv_stem(random_input)
    out = model.bn1(out)
    out = model.act1(out)
    assert out.shape == (2, 32, 112, 112)

    # Test each block, shapes found at Table 1 of EfficientNet paper
    block_act_shape = [
        (2, 16, 112, 112),
        (2, 24, 56, 56),
        (2, 24, 56, 56),
        (2, 40, 28, 28),
        (2, 40, 28, 28),
        (2, 80, 14, 14),
        (2, 80, 14, 14),
        (2, 80, 14, 14),
        (2, 112, 14, 14),
        (2, 112, 14, 14),
        (2, 112, 14, 14),
        (2, 192, 7, 7),
        (2, 192, 7, 7),
        (2, 192, 7, 7),
        (2, 192, 7, 7),
        (2, 320, 7, 7),
    ]
    for i, block in enumerate(model.blocks):
        out = block(out)
        assert out.shape == block_act_shape[i]

    out = model.conv_head(out)
    assert out.shape == (2, 1280, 7, 7)



# Copyright 2021 MosaicML. All Rights Reserved.

from typing import Any

import torch
import torch.nn as nn
import torch.nn.functional as F


class SimpleModel(nn.Module):

    def __init__(self):
        super().__init__()

        self.conv1 = nn.Conv2d(1, 16, (3, 3), padding=0)
        self.conv2 = nn.Conv2d(16, 32, (3, 3), padding=0, stride=2)
        self.bn = nn.BatchNorm2d(32)
        self.fc1 = nn.Linear(32 * 16, 32)
        self.fc2 = nn.Linear(32, 10)

    def forward(self, x: Any):
        out = self.conv1(x)
        out = F.relu(out)
        out = self.conv2(out)
        out = self.bn(out)
        out = F.relu(out)
        out = F.adaptive_avg_pool2d(out, (4, 4))
        out = torch.flatten(out, 1, -1)
        out = self.fc1(out)
        out = F.relu(out)
        return self.fc2(out)



# Copyright 2021 MosaicML. All Rights Reserved.

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Dict, Optional

import yahp as hp

from composer.core.types import JSON, Model
from composer.models.model_hparams import ModelHparams


@dataclass
class TransformerHparams(ModelHparams, ABC):
    """
    Defines the necessary hyparameters for a Transformer base module.
    """

    tokenizer_name: str = hp.optional("Model name to pull from Huggingface Model Hub.", default=None)
    pretrained_model_name: Optional[str] = hp.optional(
        doc="Pretrained model name to pull from Huggingface Model Hub.",
        default=None,
    )
    model_config: Dict[str, JSON] = hp.optional(doc="A dictionary providing a HuggingFace model configuration.",
                                                default_factory=dict)
    use_pretrained: bool = hp.optional("Whether to initialize the model with the pretrained weights.", default=False)

    def validate(self):
        if self.tokenizer_name is None:
            raise ValueError('tokenizer_name cannot be None. Enter model name to pull from HuggingFace Model Hub.')

        if self.pretrained_model_name is None and self.model_config == {}:
            raise Exception("One of pretrained_model_name or model_config needed.")

        if self.pretrained_model_name is not None and self.model_config != {}:
            raise Exception("Only one of pretrained_model_name or model_config can be provided.")

        if self.use_pretrained and self.model_config:
            raise Exception("A model cannot load pretrained weights from configuration.")

    @abstractmethod
    def initialize_object(self) -> Model:
        pass



# Copyright 2021 MosaicML. All Rights Reserved.

from composer.algorithms import functional as functional
from composer.core import Algorithm as Algorithm
from composer.core import Callback as Callback
from composer.core import Engine as Engine
from composer.core import Event as Event
from composer.core import Logger as Logger
from composer.core import State as State
from composer.core import types as types
from composer.datasets import DataloaderSpec as DataloaderSpec
from composer.trainer import Trainer as Trainer



# Copyright 2021 MosaicML. All Rights Reserved.

from dataclasses import dataclass

import yahp as hp

from composer.models.model_hparams import ModelHparams


@dataclass
class EfficientNetB0Hparams(ModelHparams):
    drop_connect_rate: float = hp.optional(
        doc="Probability of dropping a sample within a block before identity connection.",
        default=0.2,
    )

    def initialize_object(self):
        from composer.models import EfficientNetB0
        return EfficientNetB0(self)



# Copyright 2021 MosaicML. All Rights Reserved.

from dataclasses import dataclass
from typing import Optional

import yahp as hp

from composer.algorithms import AlgorithmHparams
from composer.core import Algorithm, Event, Logger, State


@dataclass
class DummyHparams(AlgorithmHparams):

    dummy_argument: str = hp.optional(doc="Placeholder dummy argument", default="default")

    def initialize_object(self):
        return Dummy(dummy_argument=self.dummy_argument)


class Dummy(Algorithm):

    def __init__(self, dummy_argument: str = 'default'):
        self.hparams = DummyHparams(dummy_argument)

    def match(self, event: Event, state: State) -> bool:
        return True

    def apply(self, event: Event, state: State, logger: Logger) -> Optional[int]:
        pass



# Copyright 2021 MosaicML. All Rights Reserved.

from dataclasses import dataclass

import yahp as hp
from torchvision import transforms
from torchvision.datasets import CIFAR10

from composer.datasets.hparams import DataloaderSpec, DatasetHparams


@dataclass
class CIFAR10DatasetHparams(DatasetHparams):
    is_train: bool = hp.required("whether to load the training or validation dataset")
    datadir: str = hp.required("data directory")
    download: bool = hp.required("whether to download the dataset, if needed")
    drop_last: bool = hp.optional("Whether to drop the last samples for the last batch", default=True)
    shuffle: bool = hp.optional("Whether to shuffle the dataset for each epoch", default=True)

    def initialize_object(self) -> DataloaderSpec:
        cifar10_mean, cifar10_std = [0.4914, 0.4822, 0.4465], [0.247, 0.243, 0.261]
        datadir = self.datadir

        if self.is_train:
            transformation = transforms.Compose([
                transforms.RandomCrop(32, padding=4),
                transforms.RandomHorizontalFlip(),
                transforms.ToTensor(),
                transforms.Normalize(mean=cifar10_mean, std=cifar10_std),
            ])
        else:
            transformation = transforms.Compose([
                transforms.ToTensor(),
                transforms.Normalize(mean=cifar10_mean, std=cifar10_std),
            ])

        return DataloaderSpec(
            dataset=CIFAR10(
                datadir,
                train=self.is_train,
                download=self.download,
                transform=transformation,
            ),
            drop_last=self.drop_last,
            shuffle=self.shuffle,
        )



# Copyright 2021 MosaicML. All Rights Reserved.

from typing import Type

from composer.loggers.logger_hparams import (BaseLoggerBackendHparams, FileLoggerBackendHparams,
                                             TQDMLoggerBackendHparams, WandBLoggerBackendHparams)

logger_registry = {
    "file": FileLoggerBackendHparams,
    "wandb": WandBLoggerBackendHparams,
    "tqdm": TQDMLoggerBackendHparams,
}


def get_logger_hparams(name: str) -> Type[BaseLoggerBackendHparams]:
    return logger_registry[name]



# Copyright 2021 MosaicML. All Rights Reserved.

from composer.callbacks.benchmarker import Benchmarker as Benchmarker
from composer.callbacks.benchmarker import BenchmarkerHparams as BenchmarkerHparams
from composer.callbacks.callback_hparams import CallbackHparams as CallbackHparams
from composer.callbacks.grad_monitor import GradMonitorHparams as GradMonitorHparams
from composer.callbacks.lr_monitor import LRMonitor as LRMonitor
from composer.callbacks.lr_monitor import LRMonitorHparams as LRMonitorHparams
from composer.callbacks.speed_monitor import SpeedMonitor as SpeedMonitor
from composer.callbacks.speed_monitor import SpeedMonitorHparams as SpeedMonitorHparams
from composer.callbacks.torch_profiler import TorchProfiler as TorchProfiler
from composer.callbacks.torch_profiler import TorchProfilerHparams as TorchProfilerHparams



# Copyright 2021 MosaicML. All Rights Reserved.

from dataclasses import asdict, dataclass

from composer.models.model_hparams import ModelHparams


@dataclass
class ResNet101Hparams(ModelHparams):

    def initialize_object(self):
        from composer.models import ResNet101
        return ResNet101(**asdict(self))



# Copyright 2021 MosaicML. All Rights Reserved.

from copy import deepcopy

import torch

from composer.core.types import DataLoader
from composer.datasets.synthetic import SyntheticDatasetHparams, SyntheticDataType
from composer.models.base import BaseMosaicModel
from composer.models.classify_mnist.mnist_hparams import MnistClassifierHparams
from composer.optim.optimizer_hparams import SGDHparams
from composer.trainer.devices.device_gpu import DeviceGPU
from composer.trainer.trainer import Trainer
from composer.trainer.trainer_hparams import TrainerHparams
from composer.utils import ensure_tuple


def get_total_loss(model: BaseMosaicModel, dataloader: DataLoader):
    with torch.no_grad():
        total_loss = 0
        for batch in dataloader:
            outputs = model(batch)
            loss = model.loss(outputs, batch=batch)
            for l in ensure_tuple(loss):
                total_loss += l.item()
        return total_loss


def train_model(mosaic_trainer_hparams: TrainerHparams, max_epochs: int = 2, run_loss_check: bool = False):
    total_dataset_size = 16
    mosaic_trainer_hparams.train_dataset = SyntheticDatasetHparams(num_classes=2,
                                                                   shape=[1, 28, 28],
                                                                   device="cpu",
                                                                   sample_pool_size=total_dataset_size,
                                                                   one_hot=False,
                                                                   drop_last=True,
                                                                   shuffle=False,
                                                                   data_type=SyntheticDataType.SEPARABLE)
    # Not used in the training loop only being set because it is required
    mosaic_trainer_hparams.val_dataset = SyntheticDatasetHparams(num_classes=2,
                                                                 shape=[1, 28, 28],
                                                                 device="cpu",
                                                                 sample_pool_size=total_dataset_size,
                                                                 one_hot=False,
                                                                 drop_last=True,
                                                                 shuffle=False)

    mosaic_trainer_hparams.model = MnistClassifierHparams(num_classes=2)
    mosaic_trainer_hparams.optimizer = SGDHparams(lr=1e-2)
    mosaic_trainer_hparams.total_batch_size = total_dataset_size
    mosaic_trainer_hparams.max_epochs = max_epochs
    # Don't validate
    mosaic_trainer_hparams.validate_every_n_epochs = max_epochs + 1

    trainer = Trainer.create_from_hparams(mosaic_trainer_hparams)

    original_model = deepcopy(trainer.state.model)
    assert isinstance(original_model, BaseMosaicModel)

    trainer.fit()

    # The original model is on the CPU so move it to GPU if needed
    if isinstance(trainer.device, DeviceGPU):
        original_model = trainer.device.module_to_device(original_model)

    if run_loss_check and trainer.state.train_dataloader:
        initial_loss = get_total_loss(original_model, trainer.state.train_dataloader)

        unwrapped_model = trainer.state.model.module
        assert isinstance(unwrapped_model, BaseMosaicModel)
        post_fit_loss = get_total_loss(unwrapped_model, trainer.state.train_dataloader)

        assert post_fit_loss < initial_loss + 1e-5



# Copyright 2021 MosaicML. All Rights Reserved.

from composer.algorithms.cutout.cutout import CutOut as CutOut
from composer.algorithms.cutout.cutout import CutOutHparams as CutOutHparams
from composer.algorithms.cutout.cutout import cutout as cutout

_name = 'CutOut'
_class_name = 'CutOut'
_functional = 'cutout'
_tldr = 'Randomly erases rectangular blocks from the image.'
_attribution = '(DeVries et al, 2017)'
_link = 'https://arxiv.org/abs/1708.04552'
_method_card = ''



# Copyright 2021 MosaicML. All Rights Reserved.

from typing import TYPE_CHECKING

from composer.models.transformer_hparams import TransformerHparams

if TYPE_CHECKING:
    from composer.models.transformer_shared import MosaicTransformer


class GPT2Hparams(TransformerHparams):
    """
    Overrides TransformerHparams to create GPT-2 specific models and configs.
    """

    def initialize_object(self) -> "MosaicTransformer":
        import transformers

        from composer.models.gpt2.model import GPT2Model
        self.validate()

        if self.model_config:
            config = transformers.GPT2Config.from_dict(self.model_config)
        elif self.pretrained_model_name is not None:
            config = transformers.GPT2Config.from_pretrained(self.pretrained_model_name)
        else:
            raise ValueError('One of pretrained_model_name or model_config needed.')

        if self.use_pretrained:
            model = transformers.AutoModelForCausalLM.from_pretrained(self.pretrained_model_name)
        else:
            model = transformers.AutoModelForCausalLM.from_config(config)  #type: ignore (thirdparty)

        return GPT2Model(
            module=model,
            config=config,  #type: ignore (thirdparty)
            tokenizer_name=self.tokenizer_name,
        )



# Copyright 2021 MosaicML. All Rights Reserved.

import logging
from typing import Any, Dict, List, Optional, Protocol, Tuple, Type

import torch

log = logging.getLogger(__name__)


class ReplacementFunction(Protocol):
    """Represents a scheme for replacing a model's modules with other modules.

    For typing reasons we represent this as a ``Protocol``, but in practice this class only describes a function.
    Replacement policies return either a replacement module, or None. Return of None means that no modifications will
    be made.

    Args:
        module (torch.nn.Module): Source module
        module_index (int): Optionally used, the i-th instance of module class.

    Returns:
        torch.nn.Module, optional: replacement module, or ``None`` to indicate no modification.
    """

    def __call__(self, module: torch.nn.Module, module_index: int) -> Optional[torch.nn.Module]:
        ...


# adapted from https://github.com/microsoft/DeepSpeed/blob/master/deepspeed/module_inject/replace_module.py#L408
def replace_module_classes(
    model: torch.nn.Module,
    policies: Dict[Any, ReplacementFunction],
    recurse_on_replacements: bool = False,
    indices: Optional[Dict[Any, int]] = None,
) -> List[Tuple[torch.nn.Module, torch.nn.Module]]:
    """ Modify model in-place by recursively applying replacement policies. Replacement policies are a mapping
    of source classes and `ReplacementFunction`.

    Examples:
        The following policy::

            policies = {
                nn.Conv2d: lambda x, idx: nn.Linear(16, 32),
                nn.MaxPool2d: lambda x, idx: nn.AvgPool2d(3, stride=2),
                nn.Linear: lambda x, idx: nn.Linear(16, 64) if x.in_features == 32 else None
            }

        will replace all convolution layers with linear layers, and all max pooling with average pooling. Linear
        layers will be optionally replaced depending on the number of input features.


    Arguments:
        module: Model to modify.
        policies: Mapping of source class to replacement function. The
            replacement may be either another module or `None`. If the latter,
            this replacement is skipped.
        recurse_on_replacements: If true, policies will be applied to any module returned
            by another policy. E.g., if one replaces a `Conv2d` with a module containing
            another `Conv2d`, this new child `Conv2d` might also be replaced. This can recurse
            infinitely if the replacement policies are not conditioned on
            module properties that change over the course of the recursion.
        indices: A dictionary mapping module types to the number of times
            they've occurred so far in the recursive traversal of
            `model` and its child modules. Allows us to pass `module_index`
            to the replacement policies, so that a policy may switch behavior
            on the i-th instance of the module_class. Note that these indices
            may not correspond to the order in which modules get called in the
            forward pass.

    Returns:
        replaced_pairs: a list of pairs of
            (original module, replacement module), reflecting the replacements
            applied to `module` and its children.

    """
    replaced_pairs = []
    indices = indices if indices is not None else {c: 0 for c in policies}
    for name, child in model.named_children():
        already_recursed = False
        child_class = child.__class__
        if child_class in policies:
            module_index = indices[child_class]
            replacement = policies[child_class](
                child,
                module_index=module_index,
            )
            indices[child_class] += 1
            if replacement is not None:
                replaced_pairs.append((child, replacement))
                if recurse_on_replacements:
                    # recurse on new child object
                    replaced_pairs += replace_module_classes(
                        replacement,
                        policies,
                        recurse_on_replacements=recurse_on_replacements,
                        indices=indices,
                    )
                already_recursed = True
                setattr(model, name, replacement)

        if not already_recursed:
            replaced_pairs += replace_module_classes(
                child,
                policies,
                recurse_on_replacements=recurse_on_replacements,
                indices=indices,
            )

    return replaced_pairs


def count_module_instances(model: torch.nn.Module, module_class: Type[torch.nn.Module]) -> int:
    """
    Counts the number of instances of module_class in the model.

    Example:
        >>> model = nn.Sequential([nn.Linear(16, 32), nn.Linear(32, 64), nn.ReLU])
        >>> count_module_instances(model, nn.Linear)
        2
        >>> count_module_instances(model, (nn.Linear, nn.ReLU))
        3

    Args:
        model (torch.nn.Module): Source model
        module_class (Type[torch.nn.Module]): module_class to count. Can also be a tuple of classes.

    Returns:
        int: The number of instances of `module_class` in `model`
    """
    count = 0
    for _, child in model.named_children():
        if isinstance(child, module_class):
            count += 1
        count += count_module_instances(child, module_class)

    return count



# Copyright 2021 MosaicML. All Rights Reserved.

from dataclasses import dataclass

from composer.callbacks.callback_hparams import CallbackHparams
from composer.core import Callback, Logger, State
from composer.utils import ensure_tuple


class LRMonitor(Callback):
    """Callback to log the learning rate.

    Learning rates for each optimizer is logged on each batch
    under the ``lr-{OPTIMIZER_NAME}/group{GROUP_NUMBER}`` key.
    """

    def __init__(self) -> None:
        super().__init__()

    def batch_end(self, state: State, logger: Logger):
        assert state.optimizers is not None, "optimizers must be defined"
        for optimizer in ensure_tuple(state.optimizers):
            lrs = [group['lr'] for group in optimizer.param_groups]
            name = optimizer.__class__.__name__
            for lr in lrs:
                for idx, lr in enumerate(lrs):
                    logger.metric_batch({f'lr-{name}/group{idx}': lr})


@dataclass
class LRMonitorHparams(CallbackHparams):
    """Parameters for the :class:`LRMonitor`.

    See the documentation for the :class:`LRMonitor`.
    """

    def initialize_object(self) -> LRMonitor:
        return LRMonitor()



# Copyright 2021 MosaicML. All Rights Reserved.

from __future__ import annotations

import logging
from dataclasses import asdict, dataclass
from typing import TYPE_CHECKING, Any, Optional, Tuple

import torch
from torchmetrics.classification.accuracy import Accuracy

from composer.algorithms import AlgorithmHparams
from composer.core import Algorithm, Event, Logger, State
from composer.core.types import Metrics, Precision
from composer.models.base import BaseMosaicModel

if TYPE_CHECKING:
    from composer.core.types import Batch, Tensors

log = logging.getLogger(__name__)


# TODO: enable for mixed precision
# TODO: enable for DDP
# TODO: enable for eval
class NoOpModelClass(BaseMosaicModel):

    def __init__(self, original_model: torch.nn.Module):
        super().__init__()
        self.parameter = torch.nn.parameter.Parameter(data=torch.tensor(0.0))
        self.zero_tensor = torch.tensor(0.0)
        try:
            # For classification
            self.num_classes = original_model.num_classes
        except AttributeError:
            pass

    def loss(self, outputs: Any, batch: Batch, *args, **kwargs) -> Tensors:
        mock_loss = self.zero_tensor + self.parameter
        return mock_loss

    def forward(self, batch: Batch) -> Tensors:
        return self.zero_tensor

    def metrics(self, train: bool) -> Metrics:
        return Accuracy()

    def validate(self, batch: Batch) -> Tuple[Any, Any]:
        raise NotImplementedError("NoOpModel not supported for eval yet.")


@dataclass
class NoOpModelHparams(AlgorithmHparams):

    def initialize_object(self) -> NoOpModel:
        return NoOpModel(**asdict(self))


class NoOpModel(Algorithm):

    def __init__(self):
        pass

    def match(self, event: Event, state: State) -> bool:
        return event == Event.INIT

    def apply(self, event: Event, state: State, logger: Logger) -> Optional[int]:
        # replace model with dummy model
        if state.precision == Precision.AMP:
            raise NotImplementedError('NoOpModel not supported for AMP Precision yet.')

        state.model = NoOpModelClass(state.model)
        log.info('Replaced model with a NoOpModel')



# Copyright 2021 MosaicML. All Rights Reserved.

from collections import defaultdict
from typing import Callable

import torch
from torch.cuda.amp.grad_scaler import GradScaler, OptState, _refresh_per_optimizer_state
from torch.optim import Optimizer

from composer.core.types import Tensor


class ClosureGradScaler(GradScaler):

    def __init__(self, ddp_reduce_scalar_and: Callable[[bool], bool], ddp_reduce_tensor_sum: Callable[[Tensor], Tensor],
                 **kwargs):
        self.ddp_reduce_scalar_and = ddp_reduce_scalar_and
        self.ddp_reduce_tensor_sum = ddp_reduce_tensor_sum
        super().__init__(**kwargs)

    def _force_scaler_ready(self, optimizer: Optimizer):
        optimizer_state = self._per_optimizer_states[id(optimizer)]
        optimizer_state["stage"] = OptState.READY

    def _empty_all_grads(self, optimizer):
        for group in optimizer.param_groups:
            for param in group["params"]:
                if param.grad is not None:
                    param.grad = None

    def _unscale_grads_and_continue(self, optimizer: Optimizer):
        if (not self._enabled):
            return True
        self._check_scale_growth_tracker("step")
        optimizer_state = self._per_optimizer_states[id(optimizer)]

        if optimizer_state["stage"] is OptState.STEPPED:
            raise RuntimeError("step() has already been called since the last update().")

        if optimizer_state["stage"] is OptState.READY:
            self.unscale_(optimizer)
        inf_detected = sum(v.item() for v in optimizer_state["found_inf_per_device"].values())
        return not inf_detected

    def step(self, optimizer: Optimizer, *args, **kwargs):
        """ Always called before the optimizer step.
        Checks if the optimizer can handle AMP closures (currently only MosaicML's SAM optimizer)
        If so, it passes an AMP-modified closure to the optimizer.
        """

        closure = kwargs["closure"]

        def _amp_closure(**kwargs):
            self._force_scaler_ready(optimizer)
            self._empty_all_grads(optimizer)

            retval: float = closure(**kwargs)

            should_continue = self._unscale_grads_and_continue(optimizer)
            should_continue = self.ddp_reduce_scalar_and(should_continue)

            return retval if should_continue else None

        return optimizer.step(closure=_amp_closure)  # type: ignore

    # Mostly copied from original grad_scaler implementation
    def update(self, new_scale=None):
        """
        Updates the scale factor.

        If any optimizer steps were skipped the scale is multiplied by ``backoff_factor``
        to reduce it. If ``growth_interval`` unskipped iterations occurred consecutively,
        the scale is multiplied by ``growth_factor`` to increase it.

        Passing ``new_scale`` sets the new scale value manually. (``new_scale`` is not
        used directly, it's used to fill GradScaler's internal scale tensor. So if
        ``new_scale`` was a tensor, later in-place changes to that tensor will not further
        affect the scale GradScaler uses internally.)

        Args:
            new_scale (float or :class:`torch.cuda.FloatTensor`, optional, default=None):  New scale factor.

        .. warning::
            :meth:`update` should only be called at the end of the iteration, after ``scaler.step(optimizer)`` has
            been invoked for all optimizers used this iteration.
        """
        if not self._enabled:
            return

        _scale, _growth_tracker = self._check_scale_growth_tracker("update")

        if new_scale is not None:
            # Accept a new user-defined scale.
            if isinstance(new_scale, float):
                self._scale.fill_(new_scale)  # type: ignore[union-attr]
            else:
                reason = "new_scale should be a float or a 1-element torch.cuda.FloatTensor with requires_grad=False."
                assert isinstance(new_scale, torch.cuda.FloatTensor), reason  # type: ignore[attr-defined]
                assert new_scale.numel() == 1, reason
                assert new_scale.requires_grad is False, reason
                self._scale.copy_(new_scale)  # type: ignore[union-attr]
        else:
            # Consume shared inf/nan data collected from optimizers to update the scale.
            # If all found_inf tensors are on the same device as self._scale, this operation is asynchronous.
            found_infs = [
                found_inf.to(device=_scale.device, non_blocking=True)
                for state in self._per_optimizer_states.values()
                for found_inf in state["found_inf_per_device"].values()
            ]

            assert len(found_infs) > 0, "No inf checks were recorded prior to update."

            found_inf_combined = found_infs[0]
            if len(found_infs) > 1:
                for i in range(1, len(found_infs)):
                    found_inf_combined += found_infs[i]

            # This is the only line changed from original grad_scaler implementation
            found_inf_combined = self.ddp_reduce_tensor_sum(found_inf_combined)

            torch._amp_update_scale_(_scale, _growth_tracker, found_inf_combined, self._growth_factor,
                                     self._backoff_factor, self._growth_interval)

        # To prepare for next iteration, clear the data collected from optimizers this iteration.
        self._per_optimizer_states = defaultdict(_refresh_per_optimizer_state)
